"""
The optimize module provides several routines to fit the parameters of a SlaterKoster model
to reproduce some given bands, usually from DFT. The optimization is done using the
minimize function from SciPy.
"""


from typing import Callable, List, Tuple
import numpy as np
from scipy.optimize import minimize
from tightbinder.models import SlaterKoster
from tightbinder.fileparse import transform_sk_coefficients


def extract_model_parameters(model: SlaterKoster) -> List[float]:
    """
    Routine to obtain the parameters of a Slater-Koster model as a list.
    
    :param model: Model whose number of parameters we want to know.
    :return: Number of parameters.
    """

    parameters = np.array([])
    for onsite_energies in model.configuration["Onsite energy"]:
        parameters = np.concatenate([parameters, onsite_energies])
    
    for distance in model.configuration["SK"].keys():
        for _, coefs in model.configuration["SK"][distance].items():
            parameters = np.concatenate([parameters, coefs])
    
    if model.configuration["Spin"]:
        parameters = np.concatenate([parameters, model.configuration["Spin-orbit coupling"]])

    return parameters


def map_parameters_to_model(model: SlaterKoster, parameters: list) -> None:
    """ 
    Routine to modify the trainable parameters of a SlaterKoster model to those
    given here. Adjustable parameters are onsite energy, slater-koster amplitudes and
    spin-orbit coupling. 
    
    :param model: SlaterKoster model to modify.
    :param parameters: List with the new parameters.
    """

    original_parameters = extract_model_parameters(model)
    expected_paramters = len(original_parameters)

    if len(parameters) != expected_paramters:
        raise ValueError(f"Invalid number of parameters provided: given {len(parameters)}, expected {expected_paramters}")

    counter = 0
    for i, onsite_energies in enumerate(model.configuration["Onsite energy"]):
        nparameters = len(onsite_energies)
        model.configuration["Onsite energy"][i] = parameters[counter:counter+nparameters]
        counter += nparameters
    
    for distance in model.configuration["SK"].keys():
        for species, coefs in model.configuration["SK"][distance].items():
            nparameters = len(coefs)
            model.configuration["SK"][distance][species] = parameters[counter:counter+nparameters]
            counter += nparameters
    
    if model.configuration["Spin"]:
        nparameters = len(model.configuration["Spin-orbit coupling"])
        model.configuration["Spin-orbit coupling"] = parameters[counter:counter+nparameters]

    model.configuration["SK amplitudes"] = model.configuration["SK"]
    transform_sk_coefficients(model.configuration)


def loss_init(model: SlaterKoster, kpoints: np.ndarray, energy: np.ndarray, penalization: np.ndarray = None, map: Callable = None) -> Callable:
    """ 
    Routine to generate the loss function (least squares) in terms of the
    parameters of a given model, evaluated over a specific k path and compared with some bands, 
    typically from DFT, to fit the model bands to 
    
    :param model: SlaterKoster model. Must implement the method solve() to evaluate
        the model.
    :param kpoints: Array of shape (nk, 3) with the kpoints.
    :param energy: Matrix (n, nk) with the energy bands we want to adjust.
    :param penalization: Matrix (n, nk) to penalize specific terms in the loss.
    :return: Loss function to optimize.
    """

    if penalization is None:
        penalization = np.ones(energy.shape)

    def loss(parameters: list) -> float:
        """
        Loss function to fit model bands to DFT bands.
        
        :param parameters: List of parameters of the subyacent model.
        :return: Loss value
        """

        if map is None:
            map_parameters_to_model(model, parameters)
        else:
            map(model, parameters)
        model.initialize_hamiltonian(find_bonds=False, verbose=False)
        results = model.solve(kpoints)
        loss_value = np.sum((results.eigen_energy - energy)**2*penalization)

        return loss_value
    
    return loss



def fit(model: SlaterKoster, kpoints: np.ndarray, energy: np.ndarray, penalization: np.ndarray = None, 
        map: Callable = None, x0: list = None, method: str = None, step: float = 1e-2) -> SlaterKoster:
    """
    Routine to fit the parameters of a SlaterKoster model to reproduce some given bands.

    :param model: SlaterKoster model to fit.
    :param kpoints: Array (nk, 3) with the kpoints where the known bands are given.
    :param energy: Matrix (n, nk) with the energy bands. n must match the dimension of the 
        hamiltonian of the model.
    :param penalization: Matrix (n, nk) to penalize specific bands or kpoints in the loss.
    :param map: Custom function to map the parameters to the model. Defaults to None.
    :param x0: Initial guess for optimization. Defaults to automatic parameter extraction if None.
    :param method: Specify optimization method. To see available options, see scipy.minimize documentation.
    :param step: Value of change of parameters when minimizing. Defaults to 1e-2.
    :return: Model with the fitted parameters.
    """

    loss: Callable = loss_init(model, kpoints, energy, penalization, map=map)
    if x0 is None:
        x0 = extract_model_parameters(model)
    optimization = minimize(loss, x0, method=method, options={'eps': step})
    fitted_parameters = optimization.x
    print(f"{optimization.message}")
    print(f"Fitted parameters: {fitted_parameters}")

    if map is None:
        map_parameters_to_model(model, fitted_parameters)
    else:
        map(model, fitted_parameters)

    return model


def read_bands_from_file(filename: str) -> Tuple[np.ndarray, np.ndarray]:
    """
    Routine to read some bands from a file.
    
    :param filename: Route to file relative to current working directory.
    :return: Array (nk, 3) with the kpoints, and matrix (n, nk) with the
        energy bands.
    """

    file = open(filename, "r")
    lines = file.readlines()
    iterator = iter(lines)
    kpoints, bands = [], []
    for line in iterator:
        kpoint = [float(number) for number in line.split()]
        energies = [float(number) for number in next(iterator).split()]

        kpoints.append(kpoint)
        bands.append(energies)

    kpoints = np.array(kpoints)
    bands = np.array(bands)

    return kpoints, bands




""" 
Introduction of disorder in a system such as vacancies
or impurities. 

All this routines are intended to be used with supercells, although
no explicit check is done on that.

There can be found routines of two types:
- Random: Unless one fixes the seed (np.seed), each call to this routines
will have a different effect upon the system it is being used on
- Listing: Instead of generating pseudorandom numbers to select the target atoms,
alternatively we can specify which atoms
"""

import numpy as np
from tightbinder.models import SlaterKoster
from tightbinder.system import System
import sys
import random
from typing import List


# ----------------------------- Random routines -----------------------------
def introduce_vacancies(system: System, probability: float = 0.5) -> System:
    """ 
    Routine to introduce vacancies in a system, i.e. to remove atoms. It is a
    statistical method, meaning that each atom in the motif has a probability of being removed
    or not. Thus each call to this method would generate a different structure.

    :param system: Instance of System class or subclass derived from it
        NB: Strictly speaking, this can be used on Crystal class as well
    :param probability: Parameter to specify probability of removing an atom;
        defaults to 0.5
    :return: Modified system
    """

    prob_array = np.random.uniform(0, 1, system.natoms)
    remaining_atoms = np.where(prob_array > probability)
    system.motif = system.motif[remaining_atoms]

    return system


def introduce_impurities(system: System, energy: float = 2, probability: float = 0.5) -> System:
    """ 
    Routine to introduce impurities in the system. These impurities are implemented
    via a change in the on-site energies of the randomly selected atoms. This routine
    randomly chooses atoms according to the specified probability, meaning that each call
    will generate a different distribution.
    NB: This routine will override ALL on-site energies for the selected atoms,
    in a multi-orbital scenario. The hopping terms remain as they were before the introduction
    of impurities.
    NB2: Current implementation requires already having initialized the Hamiltonian
    to modify the corresponding matrix elements.

    :param system: Instance of System class or subclass derived from it
    :param probability: Parameter to specify probability of selecting an atom as
        an impurity. Defaults to 0.5.
    :param energy: Value of on-site energy of impurities. Defaults to 2.
    :return: Modified system.
    """

    if system.hamiltonian is None:
        print("Error: Hamiltonian must be initialized before calling introduce_impurities routine")
        sys.exit(1)

    prob_array = np.random.uniform(0, 1, system.natoms)
    selected_atoms_as_impurities = np.where(prob_array < probability)[0]

    for i in selected_atoms_as_impurities:
        species = int(system.motif[i, 3])
        norbitals = system.norbitals[species]
        atom_interval = np.arange(norbitals*i, norbitals*(i+1))
        system.hamiltonian[0][atom_interval, atom_interval] = energy * np.ones(system.norbitals)

    return system


def amorphize(system: System, spread: float = 0.1, distribution: str = "uniform", planar: bool = False) -> System:
    """ 
    Routine to amorphize a crystalline system. This routine takes each atom and displaces
    it with respect to its original position by an amount given by a distribution. Thus
    by specifying the spread of the distribution and the distribution itself, we can
    achieve higher or lesser degrees of amorphization.
    By default it will into account the size of the supercell, so that any atom that moves outside
    of it goes to the other side.
    NB: This routine DOES NOT modify the hopping parameters according to the displacements,
    this has to be done in the class defining the system.

    :param system: Instance of System class or any derived subclass
    :param spread: Maximum distance the atoms can be displaced. Given in units of
        first neighbours (spread=1 means totally random positions). Defaults to 0.1
    :param distribution: Either "uniform" or "gaussian". By default distribution is
        always uniform, U(0, spread). If we choose gaussian, then the spread acts as the
        variance of the distribution in the radial direction N(0, spread),
        the angles are still given by uniform dist.
        Also "gaussian_cartesian", which adds gaussian noise to each cartesian component directly
    :param planar: Optional parameter to enforce atom displacement happening only in the
        plane defining a 2D system. 
    :return: Modified system.
    """

    first_neighbour_distance = system.compute_first_neighbour_distance()
    max_displacement = spread * first_neighbour_distance
    if distribution is "uniform":
        radius_array = np.random.uniform(0, max_displacement, system.natoms)
    elif distribution is "gaussian":
        radius_array = np.random.normal(0, max_displacement, system.natoms)
    elif distribution is "gaussian_cartesian":
        # To be rewritten, two return points for function are bad practice
        x = np.random.normal(0, spread, system.natoms)
        y = np.random.normal(0, spread, system.natoms)
        z = np.random.normal(0, spread, system.natoms)
        displacements = np.array([x, y, z]).T

        new_motif = np.asfarray(np.copy(system.motif))
        new_motif[:, :3] += displacements
        system.motif = new_motif

        return system
    else:
        print("Error: Incorrect distribution specified in amorphize, it has to be " +
              "either uniform or gaussian")
        sys.exit(1)
    phi_array = np.random.uniform(0, 2 * np.pi, system.natoms)
    if not planar:
        theta_array = np.random.uniform(0, np.pi, system.natoms)
    else:
        theta_array = np.ones(system.natoms) * np.pi/2

    x = radius_array * np.sin(theta_array) * np.cos(phi_array)
    y = radius_array * np.sin(theta_array) * np.sin(phi_array)
    z = radius_array * np.cos(theta_array)
    displacements = np.array([x, y, z]).T

    new_motif = np.asfarray(np.copy(system.motif))
    new_motif[:, :3] += displacements
    system.motif = new_motif

    return system


def hard_spheres(system: System, radius: float = 0.1) -> System:
    """
    Routine to treat the given model as a model of hard spheres, meaning that any two atoms that are too
    close together (distance < 2*radius), are relocated so that (distance = 2*radius). This
    is intended to be used with amorphous models with a high degree of disorder, to avoid atoms
    getting too close together.

    :param system: System to be treated as a model of hard spheres.
    :param radius: Effective radius of the hard spheres. Has to be given in the units of the motif.
    :return: System with new atomic positions.
    """

    for bond in system.bonds:
        initial_atom_index, final_atom_index, cell, neigh = bond
        initial_atom = system.motif[initial_atom_index, :3]
        final_atom = system.motif[final_atom_index, :3]
        distance = np.linalg.norm(initial_atom - final_atom)
        if distance < 2*radius:
            unit_vector = (final_atom - initial_atom)/distance
            system.motif[final_atom_index, :3] = initial_atom + unit_vector * 2 * radius

    return system


def remove_dangling_bonds(system: System) -> System:
    """
    Routine to remove dangling bonds in a system, i.e. atoms that only have one bond to
    another atom of the solid. It also removes atoms that are completely disconnected from
    the solid.

    :param system: System to remove dangling bonds and disconnected atoms.
    :return: Cleaned system.
    """

    nbonds, nremoved = 0, 0
    for atom_index in range(system.natoms):
        for initial_atom, _, _, _ in system.bonds:
            if initial_atom == atom_index:
                nbonds += 1
        if nbonds < 2:
            system.remove_atom(atom_index)

            # Update bonds with new atom indices
            for i, (initial_atom, final_atom, cell, neigh) in enumerate(system.bonds):
                if initial_atom > atom_index:
                    initial_atom -= 1
                if final_atom > atom_index:
                    final_atom -= 1
                system.bonds[i] = [initial_atom, final_atom, cell, neigh]

    return system


def alloy(system: System, *concentrations: float) -> System:
    """ 
    Routine to alloy a material with two or more chemical species. In practice this means that each atom
    is reasigned to a random chemical species, while keeping its position.
    For two chemical species, is is only necessary to specify the concentration x of the first one, Ax B1-x.
    For n species, one has to specify n-1, so that Ax1 Bx2 Cx3 ... Mxn-1.

    :param system: System to modify.
    :param concentrations: Array of length n - 1, where n is the number of species. Each
        number must be between 0 and 1, and such that the sum is <= 1.
    :return: Modified system.
    """

    if len(concentrations) != system.species - 1:
        raise ValueError("Must give nspecies - 1 concentrations")

    natoms_species = [int(system.natoms*x) for x in concentrations]
    remaining_indices = np.arange(system.natoms)
    final_species = np.zeros([system.natoms])
    for index, natoms in enumerate(natoms_species):
        indices = np.random.choice(remaining_indices, natoms, replace=False)
        final_species[indices] = index
        remaining_indices = np.delete(remaining_indices, indices)
    
    final_species[remaining_indices] = len(concentrations)
    system.motif[:, 3] = final_species

    return system


# ----------------------------- Listing routines -----------------------------
def remove_atoms(system: System, indices: List[int]) -> System:
    """ 
    Routine to remove atoms from the system motif (i.e. create vacancies) according to
    a list of indices provided.

    :param indices: List with indices of those atoms we want to remove, following the same
        ordering as those atoms in the motif
    :param system: Crystal class or subclass derived from it (typically System).
    :return: Modified system.
    """

    all_atoms = list(range(len(system.natoms)))
    remaining_atoms = [index for index in all_atoms if index not in indices]
    system.motif = system.motif[remaining_atoms]

    return system


def set_impurities(system: System, indices: List[int], energy: float = 0.1) -> System:
    """ 
    Routine to set impurities on the system on the atoms specified by the list provided.

    :param system: System class or derived subclass. Must have Hamiltonian initialized
    :param indices: List with indices of those atoms we want to transform into impurities. The indices
        are referred to the order in the motif
    :param energy: On-site energy for the impurities. Defaults to 0.1.
    :return: Modified system.
    """
    
    if system.hamiltonian is None:
        print("Error: Hamiltonian must be initialized before calling introduce_impurities routine")
        sys.exit(1)

    for i in indices:
        atom_interval = np.arange(system.norbitals*i, system.norbitals*(i+1))
        system.hamiltonian[0][atom_interval, atom_interval] = energy * np.ones(system.norbitals)

    return system


def change_species(system: SlaterKoster, indices: List[int]) -> System:
    """ 
    Routine to change the species of the atoms of the motif according to 
    the array indices, which contains the new species for each atom of the motif.
    
    :param system: SlaterKoster object
    :param indices: List of indices denoting the chemical species of each atom. Must have length equal
        to system.natoms. 
    :return: Modified system.
    """

    if(len(indices) != system.natoms):
        raise ValueError("Indices array must have the same length as system.natoms")
    if np.any(indices >= system.species):
        raise ValueError('Indices array must not contain indices equal or higher than the number of species')
    if np.any(indices < 0):
        raise ValueError('Indices must be positive numbers')

    system.motif[:, 3] = indices

    return system





"""
Topology-related routines. 

It incorporates routines to compute the Wilson loop spectra,
as well as the topological invariants derived from it. Local markers such as the
local Chern number are implemented as well.
"""

from typing import Union, Tuple
import numpy as np
import matplotlib.pyplot as plt
import sys
import scipy as sp
from .system import System
from .result import Spectrum
from matplotlib.axes import Axes
from io import TextIOWrapper

def __wilson_loop(system: System, path: list) -> np.ndarray:
    """ 
    Routine to calculate the wilson loop matrix for a system along a given path. Returns a matrix
    of size Nocc x Nocc.

    :param system: System used to compute the Wilson loop.
    :param list: List of kpoints corresponing to closed path in BZ to compute Wilson loop.
    :return: Wilson loop matrix corresponding to given closed loop.
    """

    result = system.solve(path)

    filling = system.filling
    wilson_loop = np.eye(filling)
    for i in range(len(path) - 1):
        overlap_matrix = np.dot(np.conj(np.transpose(result.eigen_states[i][:, :filling])),
                                result.eigen_states[i + 1][:, :filling])
        wilson_loop = np.dot(wilson_loop, overlap_matrix)

    overlap_matrix = np.dot(np.conj(np.transpose(result.eigen_states[i + 1][:, :filling])),
                            result.eigen_states[0][:, :filling])
    wilson_loop = np.dot(wilson_loop, overlap_matrix)

    return wilson_loop


def __check_orthogonality(matrix: np.ndarray) -> bool:
    """
    Private method to check whether the columns of a given matrix are orthonormal between them 
    (using complex scalar product).

    :param matrix: Matrix to check orthogonality.
    :return: True if orthogonal, else False.
    """
    
    epsilon = 1E-16
    difference = np.dot(np.conjugate(np.transpose(matrix)), matrix)
    difference = (np.conj(difference) * difference)
    difference[difference < epsilon] = 0
    if np.linalg.norm(difference) < epsilon:
        is_orthonormal = True
    else:
        is_orthonormal = False

    return is_orthonormal


def __generate_path(k_fixed: np.ndarray, system: System, nk: int) -> np.ndarray:
    """ 
    Routine to generate a closed path in the Brilloin zone with one of the k components fixed.

    :param k_fixed: Array corresponding to kpoint. Size (1, 3)
    :param system: System whose reciprocal lattice we use to generate a path.
    :param nk: Number of kpoints in path to generate.
    :return: Array with kpoints of path, size (nk, 3)
    """
    
    basis_vector = system.reciprocal_basis[0]
    cell_side = system.reciprocal_basis[0][0]
    path = []
    # k_list = np.linspace(-cell_side/2, cell_side/2, nk + 1)
    for i in range(nk + 1):
        # basis_vector = np.array([k_list[i], 0, 0])
        kpoint = basis_vector*i/nk + k_fixed
        # kpoint = basis_vector + k_fixed
        path.append(kpoint)

    return path


def __extract_wcc_from_wilson_loop(wilson_loop: np.ndarray) -> np.ndarray:
    """ 
    Routine to calculate the Wannier charge centres from the Wilson loop matrix.
    NOTE: We truncate the eigenvalues up to a given precision to avoid numerical error due to floating point
    when computing the midpoints.

    :param wilson_loop: Matrix of the Wilson loop.
    :return: Array with eigenvalues of Wilson loop. 
    """

    eigval = np.linalg.eigvals(wilson_loop)
    eigval = (np.angle(eigval)/np.pi).round(decimals=3)

    # eigval[eigval == -1] = 1  # Enforce interval (-1, 1], NOTE: it breaks partial polarization!

    # Extra care to ensure that the eigenvalues are not doubled, either at -1 or 1
    if np.max(np.abs(eigval)) == 1:
        one_indices = np.where(eigval == 1)[0]
        if len(one_indices) == 2:
            eigval[one_indices[0]] = -1
        elif len(one_indices) == 0:
            minus_one_indices = np.where(eigval == -1)[0]
            eigval[minus_one_indices[0]] = 1

    eigval = np.sort(eigval)

    return eigval


def calculate_wannier_centre_flow(system: System, number_of_points: int, full_BZ: bool = False, 
                                  additional_k: np.ndarray = None, nk_subpath: int = 50, 
                                  refine_mesh: bool = True, pos_tol: float = 5E-2, max_iterations: int = 20) -> np.ndarray:
    """ 
    Routine to compute the evolution of the Wannier Charge Centres (WCC) through Wilson loop calculation.
    TODO: Add parameter to toggle on/off using half/full BZ.

    :param system: System used to compute the WCC evolution.
    :param number_of_points: Number of points where the WCC are computed. 
    :param full_BZ: Boolean to toggle WCC calculation on full BZ instead of half the BZ. Useful
        for Chern number calculations (full_BZ=True). Defaults to False (for Z2 TIs).
    :param additional_k: Array corresponding to kpoints to displace the closed paths by that vector.
    :param nk_subpath: Number of points for each closed path for which we compute the Wilson loop. Defaults to 50.
    :param refine_mesh: Boolean to toggle mesh refinement. Defaults to True.
    :param pos_tol: Value of the maximum difference allowed between two consecutive WCCs. Used as criteria for mesh
        refinement.
    :param max_iterations: Maximum number of times the refinement loop is allowed to be repeated. Usually, if the
        algorithm times out it means that the Z2 calculation can't converge (due to band crossings). 
        Defaults to 20.
    :return: Array with the WCCs evaluated at each kpoint.
    """

    print("Computing Wannier centre flow...")
    if system.filling is None:
        raise ValueError('Filling must be specified to compute WCC')

    all_wcc = []
    fixed_kpoints = []

    # First generate uniform mesh in BZ
    if full_BZ:
        factor = 1
    else:
        factor = 2
    for i in range(0, number_of_points + 1):
        k_fixed = system.reciprocal_basis[1]*i/(factor*number_of_points)
        if additional_k is not None:
            k_fixed += additional_k
        fixed_kpoints.append(k_fixed)

    # Compute WCC on uniform mesh
    for k_fixed in fixed_kpoints:
        path = __generate_path(k_fixed, system, nk_subpath)
        wilson_loop = __wilson_loop(system, path)
        wcc = __extract_wcc_from_wilson_loop(wilson_loop)

        all_wcc.append(wcc)
    
    # Now check if WCC evolution is smooth. If not, refine mesh where needed
    if refine_mesh:
        refined_mesh_count = 0
        converged = False
        n_iterations = 0
        while not converged:
            print(f"iteration {n_iterations}")
            n_iterations += 1
            added_points_iteration = 0
            all_wcc_copy = np.copy(all_wcc)
            kpoints_copy = np.copy(fixed_kpoints)
            for i, wcc in enumerate(all_wcc_copy[:-1]):
                next_wcc = all_wcc_copy[i + 1]

                if __detect_wcc_position_difference(wcc, next_wcc, pos_tol):
                    new_kpoint = (kpoints_copy[i] + kpoints_copy[i + 1])/2
                    path = __generate_path(new_kpoint, system, nk_subpath)
                    wilson_loop = __wilson_loop(system, path)
                    new_wcc = __extract_wcc_from_wilson_loop(wilson_loop)

                    all_wcc.insert(i + added_points_iteration + 1, new_wcc)
                    fixed_kpoints.insert(i + added_points_iteration + 1, new_kpoint)
                    added_points_iteration += 1

            refined_mesh_count += added_points_iteration
            if (added_points_iteration == 0 or n_iterations == max_iterations):
                converged = True
        
        print(f"Invariant required computing {refined_mesh_count} extra points")

    return np.array(all_wcc)


def __detect_wcc_position_difference(wcc: np.ndarray, next_wcc: np.ndarray, pos_tol: float) -> bool:
    """ 
    Private routine to check the difference in positions of contiguous WCC. If there are not centers
    such that the difference is below the tolerance. Used for the mesh refinement algorithm in 
    calculate_wannier_centre_flow. 

    :param wcc: Array with the WCCs corresponding to one path.
    :param next_wcc: Array with the WCCs of the consecutive path.
    :param pos_tol: Maximum value allowed for the difference between WCCs of consecutive paths.
    :return: True if the difference between WCCs is higher than pos_tol. 
    """

    # Forward pass
    next_wcc_copy = np.append([next_wcc[-1] - 2], next_wcc)
    next_wcc_copy = np.append(next_wcc_copy, [next_wcc[0] + 2])
    for center in wcc:
        wcc_diff = np.abs(center - next_wcc_copy)
        if np.min(wcc_diff) > pos_tol:
            return True

    # Backward pass
    wcc_copy = np.append([wcc[-1] - 2], wcc)
    wcc_copy = np.append(wcc_copy, [wcc[0] + 2])
    for center in next_wcc:
        wcc_diff = np.abs(center - wcc_copy)
        if np.min(wcc_diff) > pos_tol:
            return True
    
    return False

# ---------------------------- Chern number ----------------------------
def calculate_polarization(wcc: np.ndarray) -> float:
    """ 
    Routine to compute the polarization from the WCC. Note that it requires computing
    the WCCs in the whole BZ, and not half BZ (which is the default calculation). 
    Useful for Chern insulators. TODO: Has to be fixed.
    
    :param wcc: Array with the WCCs for all paths.
    :return: Polarization. 
    """

    polarization = np.sum(wcc)
    return polarization


def calculate_chern_number(wcc_flow: np.ndarray) -> float:
    """ 
    Routine to compute the first Chern number from the WCC flow. Useful for Chern insulators.
    NB: This routine gives an estimate of the Chern number (non quantized), and is more precise 
    the more kpoints there are in the calculation.

    :param wcc_flow: Array with the WCCs for all paths.
    :return: Value of Chern number.
    """

    wcc_k0 = wcc_flow[0, :]
    wcc_k2pi = wcc_flow[-2, :]

    chern_number = calculate_polarization(wcc_k0) - calculate_polarization(wcc_k2pi)
    chern_number = 0 if chern_number < 1E-16 else chern_number
        
    return chern_number / 2.


# ---------------------------- Z2 invariant ----------------------------
def __find_wcc_midpoint_gap(wcc: np.ndarray) -> Tuple[float, float]:
    """ 
    Routine to find the midpoint of the biggest gap between WCCs. 
    
    :param wcc: Array with WCCs corresponding to a single path.
    :return: Position of midpoint of biggest gap among the WCCs, and index within WCC array. 
    """

    # wcc_extended = np.append(wcc, wcc[0] + 2)
    wcc_extended = np.append(wcc[-1] - 2, wcc)
    gaps = []
    for i in range(len(wcc_extended) - 1):
        gaps.append(wcc_extended[i+1] - wcc_extended[i])
    gaps = np.array(gaps).round(decimals=3)
    biggest_gap = np.max(gaps)
    position_gap = np.where(gaps == biggest_gap)[0][0]

    midpoint_gap = biggest_gap/2 + wcc_extended[position_gap]
    midpoint_gap = round(midpoint_gap, 3)
    midpoint_gap = midpoint_gap + 2 if midpoint_gap < -1 else midpoint_gap

    return midpoint_gap, position_gap


def calculate_wcc_gap_midpoints(wcc_flow: np.ndarray) -> Tuple[list, list]:
    """ 
    Routine to compute the midpoint of the biggest gap for the WCC flow, and
    their relative positions to the WCC.
    
    :param wcc_flow: Array with the WCCs for all paths.
    :return: Position and index of midpoints fort he WCCs in each path.
    """

    midpoints, positions = [], []
    for wcc in wcc_flow:
        gap, position = __find_wcc_midpoint_gap(wcc)
        midpoints.append(gap)
        positions.append(position)

    return midpoints, positions


def __directed_triangle(phi1: float, phi2: float, phi3: float) -> float:
    """ 
    Routine to determine sign of a directed triangle. Used to determine
    whether the midpoint jump over the WCC bands. Deprecated over direct counting
    of crossed WCCs.

    :param phi1: First angle.
    :param phi2: Second angle.
    :param phi3: Third angle.
    :return: Orientation of the angle.
    """

    if abs(phi1 - phi2) < 1E-14:
        g = 1
    else:
        g = np.sin(phi2 - phi1) + np.sin(phi3 - phi2) + np.sin(phi1 - phi3)
    return g


def count_crossings(wcc: np.ndarray, midpoint: float, next_midpoint: float) -> int:
    """ 
    Routine to compute the number of WCC that lie between one midpoint and the
    next one. Used to determine the Z2 invariant.
    
    :param wcc: Array with WCCs for one closed path.
    :param midpoints: Value of midpoint of the current WCC array.
    :param next_midpoints: Value of midpoint of the consecutive array of WCCs.
    :return: Number of WCCs crossed when going from the present midpoint to the next one.
    """
    
    higher_midpoint = max(midpoint, next_midpoint)
    lower_midpoint = min(midpoint, next_midpoint)
    crossings = 0
    for center in wcc:
        if lower_midpoint <= center < higher_midpoint:
            crossings += 1

    return crossings


def calculate_z2_invariant(wcc_flow: np.ndarray) -> int:
    """ 
    Routine to compute the z2 invariant from the number of jumps the
    max gap midpoint does across different WCC bands. 
    
    :param wcc_flow: Array with the WCC for all paths.
    :return: Value of Z2 invariant.
    """
    
    nk = len(wcc_flow[:, 0])
    num_crosses = 0
    for i in range(nk - 1):
        wcc = np.copy(wcc_flow[i, :])
        next_wcc = wcc_flow[i + 1, :]

        midpoint, _ = __find_wcc_midpoint_gap(wcc)
        next_midpoint, _ = __find_wcc_midpoint_gap(next_wcc)
        num_crosses += count_crossings(next_wcc, midpoint, next_midpoint)

    invariant = num_crosses % 2
    return invariant


def plot_wannier_centre_flow(wcc_flow: np.ndarray, show_midpoints: bool = True, ax: Axes = None, title: str = None,
                             fontsize: int = 10, symmetric: bool = False, full_BZ: bool = False) -> None:
    """ 
    Routine to plot the WCC evolution. 
    
    :param wcc_flow: Array with the WCCs for all paths.
    :param show_midpoints: Boolean to toggle on showing the midpoint for each WCC. Defaults to True.
    :param ax: Matplotlib Axes object to plot the WCC evolution on. Defaults to None.
    :param title: Name of the plot. Defaults to empty string.
    :param fontsize: Value to change the fontsize in the plot. Defaults to 10.
    :param symmetric: Boolean to plot the WCC over the whole BZ instead of half of it.
    :param full_BZ: Boolean to plot WCC over complete BZ. Defaults to False.
    """

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    for wcc in wcc_flow.T:
        if symmetric:
            wcc = np.concatenate((wcc[::-1][:-1], wcc), axis=0)
        ax.plot(wcc, 'ob', fillstyle='none')

    if show_midpoints:
        midpoints, _ = calculate_wcc_gap_midpoints(wcc_flow)
        if symmetric:
            midpoints = np.concatenate((midpoints[::-1][:-1], midpoints))
        ax.plot(midpoints, 'Dr', label="Midpoints")

    if title is not None:
        ax.set_title('WCC flow ' + title)
    ax.set_xlabel(r'$k_y$', fontsize=fontsize)
    ax.set_ylabel(r'$\hat{x}_n$', fontsize=fontsize)
    # ax.set_xticks([0, wcc_flow.shape[0]/2 - 1/2, wcc_flow.shape[0] - 1])
    if symmetric:
        ax.set_xticks([0, wcc_flow.shape[0] - 1, 2*wcc_flow.shape[0] - 2])
        ax.set_xticklabels([r'$-\pi$', r"0", r'$\pi$'], fontsize=fontsize)
    else:
        ax.set_xticks([0, wcc_flow.shape[0] - 1])
        if not full_BZ:
            ax.set_xticklabels([r"0", r'$\pi$'], fontsize=fontsize)
        else:
            ax.set_xticklabels([r"-$\pi$", r'$\pi$'], fontsize=fontsize)

    ax.tick_params(axis="both", labelsize=fontsize)
    ax.set_xlim([0, wcc_flow.shape[0] - 1])
    ax.set_ylim(bottom=-1, top=1)


def plot_polarization_flow(wcc_flow: np.ndarray) -> None:
    """ 
    Routine to plot the polarization flow used to calculate the Chern number. 
    
    :param wcc_flow: Array with the WCCs for all paths.
    """
    
    polarization_array = []
    for wcc in wcc_flow:
        polarization_array.append(calculate_polarization(wcc))

    plt.figure()
    plt.plot(polarization_array)


def chern_marker(system: System, results: Spectrum, ef: float = 0) -> np.ndarray:
    """
    Routine to evaluate the Chern marker at every position of the lattice.
    Must be used with systems with OBC or PBC at k=0.

    :param system: System to compute the Chern marker.
    :param results: Spectrum object with the results from the diagonalization.
    :param ef: Fermi energy, defaults to 0 (for particle-hole symmetric systems).
    :return: Array with the Chern marker evaluated at each atomic position.
    """

    
    P = np.diag((results.eigen_energy < ef).reshape(-1))
    Q = np.eye(system.basisdim) - P
    x = np.zeros([system.basisdim, system.basisdim])
    y = np.zeros([system.basisdim, system.basisdim])
    it = 0
    for n, atom in enumerate(system.motif):
        norbitals = system.norbitals[int(atom[3])]
        for i in range(norbitals):
            x[it, it] = atom[0]
            y[it, it] = atom[1]
            it += 1

    eigvec = results.eigen_states[0]

    x = eigvec.T.conj() @ x @ eigvec
    y = eigvec.T.conj() @ y @ eigvec

    PxP, PyP = P @ x @ P, P @ y @ P

    C = 2 * np.pi * 1j * (PxP @ PyP - PyP @ PxP)

    # PxQ, PyQ, QxP, QyP = P @ x @ Q, P @ y @ Q, Q @ x @ P, Q @ y @ P
    # C1 = 2 * np.pi * np.imag(QxP @ PyQ - QyP @ PxQ)
    # C2 = -2 * np.pi * np.imag(PxQ @ QyP - PyQ @ QxP)
    # C = (C1 + C2)/2
    C = eigvec @ C @ eigvec.T.conj()
    C = np.diag(C).real
    
    # Finally sum over orbitals
    # it = 0
    # trC = []
    # for n in range(system.natoms):
    #     species = int(system.motif[n, 3])
    #     norbitals = int(system.norbitals[species])
    #     summed_chern = np.sum(C[it : it + norbitals])
    #     trC.append(summed_chern)
    #     it += norbitals
    
    return C

    return C

# ---------------------------- Entanglement entropy ----------------------------
def __truncate_eigenvectors(eigenvectors: np.ndarray, sector: np.ndarray, system: System) -> np.ndarray:
    """ 
    Routine to truncate the eigenvectors according to the filling and the
    atoms that are on a specific partition, accounting for orbitals and spin structure.
    
    :param eigenvectors: Matrix with eigenvectors as columns.
    :param sector: 1d array of indices of atoms where the eigenvectors are evaluated.
    :param system: System used (necessary to extract orbital information and filling).
    :return: Eigenvectors with entries corresponding only to the specified sector,
        and truncated with the filling.
    """

    orbitals = []
    counter = 0
    for n in range(system.natoms):
        norb = system.norbitals[int(system.motif[n, 3])]
        if n in sector:
            for _ in range(0, norb):
                orbitals.append(counter)
                counter += 1
        else:
            counter += norb

    truncated_eigenvectors = eigenvectors[orbitals]
    truncated_eigenvectors = truncated_eigenvectors[:, :system.filling]

    return truncated_eigenvectors


def __density_matrix(truncated_eigenvectors: np.ndarray) -> np.ndarray:
    """ 
    Private routine to calculate the one-particle reduced density matrix for the half-system. 
    
    :param truncated_eigenvectors: Matrix with truncated eigenvectors.
    :return: Density matrix.
    """
    
    
    density_matrix = np.dot(truncated_eigenvectors, np.conj(truncated_eigenvectors.T))

    return density_matrix


def specify_partition_plane(system: System, plane: list) -> np.ndarray:
    """ 
    Routine to specify a real-space partition of the atoms of the system based on a given
    plane.
    
    :param system: System whose atoms we want to separate into two partitions.
    :param plane: List with coefficients of the plane generating the partition. 
        Array [A,B,C,D] <-> Ax + By + Cz = D
    :return: Array with indices of atoms on a side. """

    sector = []
    plane_normal_vector = plane[:3]
    plane_coefficient = plane[3]

    for i, atom in enumerate(system.motif):
        atom_position = atom[:3]
        side = np.dot(atom_position, plane_normal_vector) - plane_coefficient
        if np.sign(side) == -1:
            sector.append(i)

    if not sector or len(sector) == len(system.motif):
        print("Error: Could not find atoms with specified plane, exiting...")
        sys.exit(1)

    return np.array(sector)


def specify_partition_shape(system: System, shape: str, center: np.ndarray, r: float) -> np.ndarray:
    """ 
    Routine to determine which atoms are inside of a partition of the system.
    
    :param system: System object to extract the motif positions.
    :param shape: Either 'square' or 'circle'. Shape of the partition we are considering.
    :param center: Array with the coordinates of the center of the shape.
    :param r: Radius of the shape; for a square it corresponds to its side.
    :return partition: Indices of atoms referred to the motif ordering which are
        inside the specified partition.
    :return: Array with indices of atoms on a side. """

    """ TODO specify_partition_shape: 'square' shape not implemented """
    available_shapes = ["square", "circle"]
    if shape not in available_shapes:
        raise ValueError("shape must be either square or circle")

    partition = []
    for n, atom in enumerate(system.motif):
        atom_position = atom[:3]
        if np.linalg.norm(atom_position - np.array(center)) < r:
            partition.append(n)

    return np.array(partition)


def entanglement_spectrum(system: System, partition: np.ndarray, kpoints: np.ndarray = None) -> np.ndarray:
    """ 
    Routine to obtain the eigenvalues from the correlation/density matrix, which is directly
    related with the entangled Hamiltonian. Should be computable for both PBC and OBC.
    
    :param system: System where we want to compute the the entanglement spectrum.
    :param partition: Array with the indices of the atoms in the partition where we want to 
        compute the entanglement spectrum.
    :param kpoints: Array of kpoints where we evaluate the entanglement spectrum. Defaults to None
        which is [0., 0., 0.] for both OBC and PBC.
    :return: Entanglement spectrum evaluated at each kpoint (if given).
    """
    
    if kpoints is None and system.boundary != "OBC":
        print('PBC: Defaulting kpoints to origin...')
        kpoints = [[0., 0., 0.]]
    if system.boundary == "OBC":
        if kpoints is not None:
            print("Warning: kpoints argument given but system uses OBC")
            print("OBC: Defaulting kpoints to origin...")
        kpoints = [[0., 0., 0.]]

    print("Computing entanglement spectrum...")
    print("Diagonalizing system...")
    nk = len(kpoints)
    results = system.solve(kpoints)

    basisdim = np.sum([system.norbitals[int(system.motif[atom, 3])] for atom in partition])
    spectrum = np.zeros([basisdim, nk])
    print("Computing density matrices...")
    for i in range(len(kpoints)):
        truncated_eigenvectors = __truncate_eigenvectors(results.eigen_states[i],
                                                         partition, system)

        density_matrix = __density_matrix(truncated_eigenvectors)
        eigenvalues, _ = np.linalg.eigh(density_matrix)

        spectrum[:, i] = eigenvalues

    return spectrum


def entanglement_entropy(spectrum: np.ndarray) -> float:
    """ 
    TODO entanglement_entropy fix log(0). 
    Routine to compute the entanglement entropy from the entanglement spectrum.

    :param spectrum: Array with the entanglement spectrum.
    :return: Value of entanglement entropy.
    """
    
    entropy = 0
    for eigval in spectrum.T:
        entropy -= np.dot(eigval, np.log(eigval)) + np.dot(1 - eigval, np.log(1 - eigval))
    entropy /= len(spectrum.T)

    return entropy


def write_entanglement_spectrum_to_file(spectrum: np.ndarray, file: TextIOWrapper, n: int = None, shift: int = 0) -> None:
    """ 
    Routine to write the entanglement spectrum to a text file. If n is given,
    only n eigenvalues are written to file.
    The algorithm is as follows: first we search for the eigenvalue that is closer to 0.5, and then
    we write the n/2 eigenvalues to its left and right. 
    
    :param spectrum: Array with the entanglement spectrum.
    :param file: Pointer to the file to write the entanglement spectrum.
    """

    print("Writing entanglement spectrum to file...")
    eigenvalues = np.sort(spectrum.reshape(-1,))
    if n > len(eigenvalues):
        print("Error: n is bigger than overall number of eigenvalues")
        sys.exit(1)
    if n is None:
        n = len(eigenvalues)

    for i, eigval in enumerate(eigenvalues):
        if eigval <= 0.5 < eigenvalues[i + 1]:
            central_index = i
            break

    with open(file, "w") as textfile:
        for i in range(central_index - n//2 + shift, central_index + n//2 + shift):
            textfile.write(f"{eigenvalues[i]}\n")


def plot_entanglement_spectrum(spectrum: np.ndarray, system: System, ax: Axes = None,
                               fontsize: int = 10, title: str = None, markersize: int = 5, color: str = "b") -> None:
    """ 
    Routine to plot the entanglement spectrum as a function of k.
    CAREFUL: This routine is not made to deal with any set of kpoints, rather it is intended
    to be used with a set of kpoints along a given line (1D). 
    
    :param spectrum: Array with the entanglement spectrum.
    :param system: System where we compute the entanglement spectrum.
    :param ax: Matplotlib Axes object to plot in. Defaults to None.
    :param fontsize: Value to change the fontsize of the plot. Defaults to 10.
    :param title: Title of the plot. Defaults to empty string.
    :param markersize: Size of the markers of the plot. Defaults to 5.
    :param color: Color of the markers. Defaults to blue.
    """
    
    if ax is None:
        fig = plt.figure(figsize=(5, 5))
        ax = fig.add_subplot(111)

    if system.boundary == "OBC" or len(spectrum.shape) == 1 or spectrum.shape[1] == 1:
        ax.plot(spectrum, 'o', c=color, markersize=markersize)
        ax.set_xlim(0, len(spectrum))
        ax.set_xlabel(r"$n$", fontsize=fontsize)
    else:
        for entanglement_band in spectrum:
            ax.plot(entanglement_band, 'o', c=color)
        ax.set_xlim(0, len(spectrum.T) - 1)
        ax.set_xticks([0, len(spectrum.T)/2 - 1/2, len(spectrum.T) - 1])
        ax.set_xticklabels([r"-$\pi/a$", "0", r"$\pi/a$"], fontsize=fontsize)
        ax.set_xlabel(r"$k_y$", fontsize=fontsize)

    if title is not None:
        ax.set_title(title, fontsize=fontsize)
    ax.tick_params(axis='both', labelsize=fontsize, direction="in")

    ax.set_ylim(0, 1)
    ax.set_yticks([0, 0.5, 1])
    # ax.set_title(f"Entanglement spectrum of {system.system_name}", fontsize=15)
    
    ax.set_ylabel(rf"$\xi_n$", fontsize=fontsize)



# ---------------------- Quantum geometry ----------------------
def berry_curvature(model: System, nk: int, band_index: int = 0) -> Tuple[np.ndarray, np.ndarray]:
    """
    Calculation of the Berry curvature of a single band using the Fukui method.
    Only valid for 2D systems.
    Note that this routine is only valid for single bands without degeneracies, for
    non-abelian Berry curvatures one should use the generalized method.

    :param model: System from which we want to obtain the Berry curvature.
    :param nk: Number of k points along one axis of the BZ.
    :param band_index:
    :returns: List of kpoints and field strenght of Berry connection.
    """

    if model.ndim != 2:
        raise AttributeError("Model must be two-dimensional.")
    
    kpoints = model.brillouin_zone_mesh([nk, nk])
    
    unit_displacements = []
    for i in range(model.ndim):
        unit_displacements.append(model.reciprocal_basis[i]/nk)

    curvature = []
    for kpoint in kpoints:
        local_kpoints = [kpoint, kpoint + unit_displacements[0], 
                         kpoint + unit_displacements[0] + unit_displacements[1], 
                         kpoint + unit_displacements[1], kpoint]
        eigvec = model.solve(local_kpoints).eigen_states

        links = []
        for i in range(len(local_kpoints) - 1):
            link = np.vdot(eigvec[i][:, band_index], eigvec[i + 1][:, band_index])
            link /= np.abs(link)
            links.append(link)

        strength = -1j * np.log(np.prod(links))
        curvature.append(strength)


    return kpoints, np.array(curvature)















"""
Definition of Result class to handle diagonalization results from Hamiltonian and observable calculation.
"""

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.cm import get_cmap
from matplotlib.collections import LineCollection
import numpy as np
import math
import sys
from .utils import condense_vector, scale_array
from .system import System
from typing import List, Union

class Spectrum:
    """
    The Spectrum class is designed to store the results from the Hamiltonian diagonalization, 
    and to perform manipulations on the eigenvectors and eigenvalues and extract information
    about the system.

    :param eigen_energy: Array (n, nk) with the energies for each kpoint stored as columns.
    :param eigen_states: Array (nk, n, n) with the eigenvectors for each kpoint.
    :param kpoints: Array with kpoints where the Bloch Hamiltonian was evaluated.
    :param system: System used in the calculation.
    """

    def __init__(self, eigen_energy: np.ndarray = None, eigen_states: np.ndarray = None, kpoints: np.ndarray = None, 
                system: System = None) -> None:

        self.eigen_energy = eigen_energy
        self.eigen_states = eigen_states
        self.kpoints = kpoints
        self.system = system

    def __simplify_kpoints(self) -> np.ndarray:
        """ 
        Routine to reduce the k-points of the mesh from 3d to the corresponding dimension,
        for later graphical representation
        DEPRECATED 
        """

        dimension = self.configuration['Dimensionality']
        new_kpoints = np.zeros([len(self.kpoints), dimension])
        non_zero_index = np.nonzero(self.kpoints[0])
        if dimension == 1:
            for i, kpoint in enumerate(self.kpoints):
                new_kpoints[i, :] = kpoint[non_zero_index]

        elif dimension == 2:
            pass

        else:
            pass

        return new_kpoints

    def plot_bands(self, title: str = '', ax: Axes = None) -> None:
        """ 
        Method to plot bands from diagonalization in the whole Brillouin zone.

        :param title: Title for plot. Defaults to empty string.
        :param ax: Axes object. Defaults to None.
        """

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")

        ax.view_init(0, 0)

        nk = len(self.kpoints[:, 0])
        size = int(math.sqrt(nk))
        x = self.kpoints[:, 0].reshape(size, size)
        y = self.kpoints[:, 1].reshape(size, size)
        z = self.kpoints[:, 2].reshape(size, size)
        for i in range(len(self.eigen_energy[:, 0])):
            en = self.eigen_energy[i, :].reshape(size, size)
            ax.plot_surface(x, y, en)

        ax.set_title(title + 'band structure')
        ax.set_xlabel(r'$k_x$ ($A^{-1}$)')
        ax.set_ylabel(r'$k_y$ ($A^{-1}$)')


    def plot_along_path(self, labels: List[str], title: str = '', edge_states: bool = False, rescale: bool = True,
                        ax: Axes = None, e_values: List[float] = [], fontsize: float = 10, linewidth: float = 2,
                        edgecolor = "red") -> None:
        """ 
        Method to plot the bands along a path in reciprocal space, normally along high symmetry points.

        :param labels: Labels of the High Symmetry Points of the path.
        :param title: Title of the plot. Defaults to empty string.
        :param edge_states: Boolean parameter to toggle on or off edge bands in a different color. 
            Edge bands are defined as those immediately above and below the Fermi level. Defaults to False.
        :param rescale: Boolean to rescale the energy spectrum to the Fermi energy. I.e. highest occupied
            state has energy zero. Defaults to True.
        :param ax: Axes object from matplotlib to plot bands there. Useful for figures with subplots.
        :param e_values: List with two values, [e_min, e_max] to show bands only in that energy range.
        :param linewidth: Linewidth. Defaults to 2.
        :param fontsize: Adjusts size of lines and text.
        :param edgecolor: Color of the edge bands if edge_states=True.
        """

        # First get distances of relevant high symmetry points
        points = [self.system.high_symmetry_points[point] for point in labels]
        path_distances = [np.linalg.norm(points[i + 1] - points[i]) for i in range(len(points) - 1)]
        cummulative_path_weights = [0] 
        cummulative_path_weights += [np.sum(path_distances[:i + 1])/np.sum(path_distances) for i in range(len(labels) - 1)]

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        if e_values and len(e_values) != 2:
            raise ValueError("e_values must be [e_min, e_max]")

        if rescale:
            self.rescale_bands_to_fermi_level()

        nk = len(self.kpoints)
        x_points = np.arange(0, nk, dtype=float)
        x_ticks = []
        number_of_paths = len(labels) - 1

        label_points = []
        for n, label in enumerate(labels):
            xpos = (nk - 1)*cummulative_path_weights[n]
            x_ticks.append(xpos)
            factor = int((nk - 1)/number_of_paths)
                    
        x_points = [0]
        for n, point in enumerate(x_ticks[:-1]):
            next_point = x_ticks[n + 1]
            path_points = np.linspace(point, next_point, factor + 1, endpoint=True)
            x_points += list(path_points[1:])
                
        for eigen_energy_k in self.eigen_energy:
            ax.plot(x_points, eigen_energy_k, 'k-', linewidth=linewidth)

        if edge_states and self.system.filling is not None:
            edge_states_indices = [int(self.system.filling) + i for i in range(-2, 2)]
            for state in edge_states_indices:
                ax.plot(x_points, self.eigen_energy[state], '-', linewidth=linewidth, c=edgecolor)

        ax.set_xticks(x_ticks)
        ax.set_xticklabels(labels, fontsize=fontsize)
        ax.set_ylabel(r'$E$ (eV)', fontsize=fontsize)
        ax.tick_params('y', labelsize=fontsize)
        if title != '':
            ax.set_title(title + " band structure", fontsize=fontsize)
        ax.set_xlim([min(x_points), max(x_points)])
        if e_values:
            ax.yaxis.set_ticks(np.arange(np.round(e_values[0]), np.round(e_values[1]) + 1, 1))
            ax.set_ylim(e_values)

        ax.grid(linestyle='dashed', axis="x")

    def plot_bands_w_atomic_occupation(self, labels, atom_indices, title='', rescale=True, 
                                       ax=None, e_values=[], fontsize=10):
    
        """ 
        Method to plot bands as function of k, but also as a colormap to show edge occupation of 
        each state. 

        :param labels: Labels of the High Symmetry Points of the path.
        :param atom_indices: List of indices of atoms where we want to measure the occupation.
        :param title: Title of the plot. Defaults to empty string.
        :param edge_states: Boolean parameter to toggle on or off edge bands in a different color. 
            Edge bands are defined as those immediately above and below the Fermi level. Defaults to False.
        :param rescale: Boolean to rescale the energy spectrum to the Fermi energy. I.e. highest occupied
            state has energy zero. Defaults to True.
        :param ax: Axes object from matplotlib to plot bands there. Useful for figures with subplots.
        :param e_values: List with two values, [e_min, e_max] to show bands only in that energy range.
        :param fontsize: Adjusts size of lines and text.
        """

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        if e_values and len(e_values) != 2:
            raise ValueError("e_values must be [e_min, e_max]")

        if rescale:
            self.rescale_bands_to_fermi_level()

        nk = len(self.kpoints)
        x_points = np.arange(0, nk)
        x_ticks = []
        number_of_paths = len(labels) - 1

        for n, _ in enumerate(labels):
            xpos = (nk - 1)/number_of_paths*n
            x_ticks.append(xpos)
        
        # Sort by increasing occupation to avoid visual artifacts when plotting
        occupation_matrix = self.calculate_occupation(atom_indices)

        colormap = "viridis"
        cmap = get_cmap(colormap)
        segments = []
        colors = []
        for i, eigen_energy_k in enumerate(self.eigen_energy):
            for j in x_points[:-1]:
                segments.append([(x_points[j], eigen_energy_k[j]), (x_points[j + 1], eigen_energy_k[j + 1])])
                colors.append(cmap(occupation_matrix[i, j]))

        lines = LineCollection(segments[::-1], colors=colors[::-1], linewidth=2)
        ax.add_collection(lines)
        print(f"Max occupation: {np.max(occupation_matrix)}")

        cbar = plt.colorbar(lines, ax=ax, extend="both")
        cbar.set_label("Edge occupation", fontsize=fontsize)
        lines.set_cmap(colormap)
        cbar.ax.tick_params(labelsize=fontsize) 

        ax.set_xticks(x_ticks)
        for i, label in enumerate(labels):
            if label == "G":
                labels[i] = r"$\Gamma$"
        ax.set_xticklabels(labels, fontsize=fontsize)
        ax.set_ylabel(r'$\varepsilon$ (eV)', fontsize=fontsize)
        ax.tick_params('y', labelsize=fontsize)
        if title != '':
            ax.set_title(title + " band structure", fontsize=fontsize)
        ax.set_xlim([min(x_points), max(x_points)])
        if e_values:
            ax.yaxis.set_ticks(np.arange(np.round(e_values[0]), np.round(e_values[1]) + 1, 1))
            ax.set_ylim(e_values)
        

        # ax.grid(linestyle='--')

    def plot_spectrum(self, title: str = '',  ax: Axes = None, fontsize: int = 10) -> None:
        """ 
        Routine to plot all the eigenvalues coming from the Bloch Hamiltonian diagonalization
        in an ordered way.
        Specially suitable for open systems, although it can be used for periodic systems as well.

        :param title: Optional title for the plot.
        :param ax: Matplotlib Axes object to the plot.
        :param fontsize: Fontsize for plot labels. Defaults to 10.
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)

        all_eigenvalues = self.eigen_energy.reshape(-1, 1)
        all_eigenvalues = np.sort(all_eigenvalues)

        ax.plot(all_eigenvalues, 'g+')
        if title != '':
            ax.set_title(f"Spectrum of {title}", fontsize=fontsize)
        else:
            ax.set_title("Spectrum", fontsize=fontsize)
        ax.set_ylabel(r"$\varepsilon (eV)$", fontsize=fontsize)
        ax.set_xlabel("n", fontsize=fontsize)
        ax.tick_params('both', labelsize=fontsize)

    def write_bands_to_file(self, file: str):
        """ TODO write_bands_to_file"""
        pass

    def write_states_to_file(self, file: str):
        """ TODO write_states_to_file"""
        pass

    # --------------- Utilities ---------------
    @staticmethod
    def is_edge_state(state: np.ndarray, norbitals: int, edge_indices: List[int], penetration: float = 0.1) -> bool:
        """
        Routine to check if an eigenstate can be regarded as edge state (as in localized at the edge).
        TODO: norbitals should be a list with the orbitals for all species.

        :param state: Array with coefficients of eigenstate.
        :param norbitals: Number of orbitals per atom. NB: Must be changed to list.
        :param edge_indices: List with indices of atoms in the edge.
        :param penetration: Maximum probability allowed for the state to be in non-edge atoms.
            Defaults to 0.1
        :return: True or False.
        """
        
        orbital_probabilities = np.abs(state) ** 2
        amplitudes = condense_vector(orbital_probabilities, norbitals)
        edge_density = np.sum(amplitudes[edge_indices])

        is_edge_state = False
        if edge_density > (1 - penetration):
            is_edge_state = True

        return is_edge_state

    def identify_edge_states(self, system: System, penetration: float = 0.1) -> List[int]:
        """ 
        Routine to identify edge states according to a given penetration parameter. To
        be used with OBC only.

        :param system: System of interest.
        :param penetration: Maximum probability allowed for the states to be in non-edge atoms.
            Defaults to 0.1
        :return: List of states located at the edge.
        """

        if self.eigen_states.shape[0] != 1:
            print("Error: identify_edge_states can only be used with OBC, exiting...")
            sys.exit(1)
        edge_indices = system.identify_edges()
        edge_indices = system.find_lowest_coordination_atoms()
        norbitals = len(self.eigen_states[0][:, 0])//len(system.motif)
        edge_states = []
        for n, state in enumerate(self.eigen_states[0].T):
            if self.is_edge_state(state, norbitals, edge_indices, penetration):
                edge_states.append(n)

        return edge_states

    def calculate_fermi_energy(self, filling: int, bottom: bool = True) -> float:
        """ 
        Routine to compute the Fermi level energy according to the given filling .
        Fermi energy is set systematically in-between the last occupied state and the first
        unoccupied state (midgap for insulators and approximately last filled for metals).
        
        :param filling: Total number of electrons in the system.
        :param bottom: Whether to set Fermi energy at maximum of valence band. Defaults to True.
        :return: Value of Fermi energy.
        """
        
        filling = int(filling*self.eigen_energy.shape[1])
        all_energies = self.eigen_energy.reshape(-1)
        all_energies = np.sort(all_energies)
        if bottom:
            fermi_energy = all_energies[filling - 1]
        else:
            fermi_energy = (all_energies[filling - 1] + all_energies[filling])/2

        return fermi_energy

    def rescale_bands_to_fermi_level(self) -> None:
        """ 
        Routine to set the Fermi energy to zero. 
        """
        
        fermi_energy = self.calculate_fermi_energy(self.system.filling)
        self.eigen_energy -= fermi_energy

    def calculate_gap(self) -> float:
        """ 
        Routine to compute the gap of a material based on its Fermi level/filling. 
        
        :return: Gap value.
        """

        filling = self.system.filling
        
        valence_band = self.eigen_energy[filling - 1, :]
        conduction_band = self.eigen_energy[filling, :]
        gap = np.min(conduction_band - valence_band)

        return gap

    def calculate_ipr(self) -> List[float]:
        """ 
        Method to compute the inverse participation ratio (IPR) from the eigenvectors.

        :return: list with all ipr for all states 
        """
        
        eigen_states = np.copy(self.eigen_states.reshape(self.system.basisdim, -1).T)
        ipr_array = []
        for eigenvector in eigen_states:
            state = State(eigenvector, self.system)
            ipr_array.append(state.compute_ipr())

        return ipr_array

    def calculate_average_ipr(self, states: Union[np.ndarray, list]) -> List[float]:
        """ 
        Method to compute the average IPR associated to a group of states.
        
        :param states: list with the eigenvectors of the states.
        :return: Average IPR of the given states. 
        """
        
        average_ipr = 0
        for eigenvector in states:
            state = State(eigenvector, self.system)
            average_ipr += state.compute_ipr()

        average_ipr /= len(states)
        return average_ipr

    def calculate_occupation(self, atom_indices: List[int]) -> np.ndarray:
        """ 
        Method to compute the occupation for all the states in the spectrum
        on the specified atoms.

        :param atom_indices: List with the indices of the atoms where we want to obtain the occupation.
        :return: Matrix with the occupations of the states, with the shape of eigen_energy.
        """
        
        occupations = np.zeros(self.eigen_energy.shape)
        for kIndex, k_eigenstates in enumerate(self.eigen_states):
            for i, eigenstate in enumerate(k_eigenstates.T):
                state = State(eigenstate, self.system)
                occupations[i, kIndex] = state.compute_occupation(atom_indices)

        return occupations

    @staticmethod
    def plot_quantity(array: Union[list, np.ndarray], name: str = None, sort: bool = False, ax: Axes = None) -> None:
        """ 
        Method to plot an array of some quantity for each state
        in the spectrum as a bar plot 
        
        :param array: Array with values to be plotted.
        :param name: Optional title for the plot. Defaults to None.
        :param sort: If True, sorts the values of array. Defaults to False.
        :param ax: Axes object to plot in.
        """
        
        x = np.arange(0, len(array))

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)

        if sort:
            array = np.sort(array)
        ax.bar(x, array, width=1)
        ax.set_title(fr"{name}")
        ax.set_ylabel(f"{name}")
        ax.set_xlabel("States")


class State:
    """
    The State class is defined to separate part of the functionality of the Spectrum class
    to the eigenstates themselves.
    
    :param eigenvector: Array with coefficients of eigenvector.
    :param system: System object.
    """

    def __init__(self, eigenvector: np.ndarray, system: System):
        self.eigenvector = eigenvector
        self.motif       = system.motif
        self.norbitals   = system.norbitals
        self.hoppings    = system.bonds
        self.system      = system
        self.amplitude   = self.atomic_amplitude()

    def atomic_amplitude(self) -> None:
        """ 
        Method to obtain the probability amplitude corresponding to each atom.
        :return: Array with the probability of finding the electron on each atom (i.e. ordered by atoms like motif).
        """

        amplitudes = np.abs(self.eigenvector) ** 2
        reduced_vector = []
        counter = 0
        for atom_index in range(self.motif.shape[0]):
            species = self.motif[atom_index, 3]
            orbitals = self.norbitals[int(species)]
            atom_amplitude = amplitudes[counter:counter + orbitals]
            reduced_vector.append(np.sum(atom_amplitude))
            counter += orbitals

        return np.array(reduced_vector)

    def plot_amplitude(self, ax: Axes = None, title: str = None, linewidth: int = 1, bonds: bool = True, factor: int = 200):
        """ 
        Method to plot the atomic amplitude of the state on top of the crystalline positions.
        
        :param ax: Axes object to plot the amplitude. Defaults to None.
        :param title: Plot title. Defaults to None.
        :param linewidth: Change width of bonds if present. Defaults to 1.
        :param bonds: Toggle on/off bonds between atoms. Defaults to True.
        :param factor: To scale the amplitude up or down.
        """
        
        amplitude = np.array(self.atomic_amplitude())
        scaled_amplitude = amplitude/np.max(amplitude)
        size_adjustment = len(amplitude)/2 - np.power((1 - scaled_amplitude), 4)
        scaled_amplitude[scaled_amplitude < 1/len(scaled_amplitude)] = 0

        if ax is None:
            fig = plt.figure(figsize=(5, 6))
            ax = fig.add_subplot(111)

        atoms = np.array(self.motif)[:, :3]
        if bonds:
            self.system.plot_wireframe(ax=ax, linewidth=linewidth)

        ax.scatter(atoms[:, 0], atoms[:, 1], c=scaled_amplitude,
                   cmap="plasma", alpha=1, s=scaled_amplitude*factor)
        ax.set_xlim([np.min(atoms[:, 0]), np.max(atoms[:, 0])])
        ax.set_ylim([np.min(atoms[:, 1]), np.max(atoms[:, 1])])
        ax.set_xticks([0, 29])
        ax.set_yticks([0, 29])
        if title is not None:
            ax.set_title(title)
        ax.axis('off')
        ax.axis('tight')
        ax.axis('equal')

    def compute_ipr(self) -> float:
        """ 
        Method to compute the inverse participation ratio (IPR) for the state. 
        
        :return: IPR of state.
        """
        
        squared_amplitude = self.amplitude ** 2
        ipr = np.sum(squared_amplitude)

        return ipr

    def compute_occupation(self, atoms_indices: List[int]) -> float:
        """ 
        Method to compute which percentage of the atomic amplitude is located
        at the given list of atoms. 
        
        :param atom_indices: List of indices of atoms where we want to compute the occupation.
        """
        
        occupation = np.sum(self.amplitude[atoms_indices])
        return occupation

    def compute_spin_projection(self, axis: str = "z"):
        """ 
        TODO compute_spin_projection.
        Method to compute the expected value of any of the three spin operators Sx, Sy or Sz.
        All calculations are done supposing that the original atomic basis for the tight-binding is written
        using the z projection of the spin (if the calculation is spinful).
        
        :param axis: 'x', 'y' or 'z'.
        :return: Expected value of spin along the given direction.
        """

        axis_list = ["x", "y", "z"]
        if axis not in axis_list:
            raise KeyError("Axis must be x, y or z")

        pass








"""
Miscellaneous routines used in the other modules.
"""


from typing import Callable, Union
import numpy as np
from scipy.spatial import Delaunay
import hashlib


def generate_basis_combinations(ndim: int) -> np.ndarray:
    """ 
    Routine to generate the coefficients corresponding to linear combinations of
    some basis vectors. Possible coefficients for each vector are -1, 0 and 1.

    :param ndim: Dimension of the basis.
    :return: Matrix where each row are the coefficients for a given
        combination of the basis vectors, i.e. [c1, c2, ...] such that
        v = c1*basis_1 + c2*basis_2 + ... 
    """

    mesh_points = []
    if ndim == 0:
        return mesh_points
    for i in range(ndim):
        mesh_points.append(list(range(-1, 2)))
    mesh_points = np.array(np.meshgrid(*mesh_points)).T.reshape(-1, ndim)

    return mesh_points


def condense_vector(vector: Union[list, np.ndarray], step: int) -> np.ndarray:
    """ 
    Routine to reduce the dimensionality of a given vector by summing each consecutive n (step) numbers.

    :param vector: Array to be reduced
    :param step: Number of components to sum
    :returns: Condensed array
    """
    
    if (len(vector) % step) != 0:
        raise ArithmeticError("Step must divide vector length")
    if step > len(vector):
        raise ArithmeticError("Step can not be larger than vector length")

    reduced_vector = []
    for i in range(len(vector)//step):
        atom_amplitudes = vector[i*step:(i+1)*step]
        reduced_vector.append(np.sum(atom_amplitudes))

    return np.array(reduced_vector)


def scale_array(array: Union[list, np.ndarray], factor: int = 10) -> np.ndarray:
    """ 
    Routine to scale a vector by a factor max_value/max(vector), where max_value is the new maximum value.

    :param array: Array to scale.
    :param factor: Factor to multiply the array. Defaults to 10
    :return: Scaled array.
    """

    n = len(array)
    array = np.array(array) * n * factor

    return array


def pretty_print_dictionary(d: dict, indent: int = 0) -> None:
    """
    Routine to pretty print the contents of a dictionary.
    Note that this is mainly used to indent recursively all dict. contents, so by
    default it should be left at zero.

    :param d: Dictionary to pretty print
    :param indent: Number of tabs to indent the contents of the dictionary.
    """

    for key, value in d.items():
        print('\t' * indent + str(key))
        if isinstance(value, dict):
            pretty_print_dictionary(value, indent+1)
        else:
            print('\t' * (indent+1) + str(value))


def overrides(interface_class: type) -> Callable:
    """
    Decorator to print whether a method is being overwritten or not.

    :param interface_class: Base class where the method to be overwritten is present.
    :return: Decorator.
    """
    def overrider(method):
        assert(method.__name__ in dir(interface_class))
        return method
    return overrider


def alpha_shape_2d(points: np.ndarray, alpha: float, only_outer: bool = True):
    """
    Compute the alpha shape (concave hull) of a set of 2D points.
    From: https://stackoverflow.com/questions/50549128/boundary-enclosing-a-given-set-of-points

    :param points: np.array of shape (n, 2) points.
    :param alpha: alpha value.
    :param only_outer: boolean value to specify if we keep only the outer border
        or also inner edges.
    :return: set of (i,j) pairs representing edges of the alpha-shape. (i,j) are
        the indices in the points array.
    """
    assert points.shape[0] > 3, "Need at least four points"

    def add_edge(edges, i, j):
        """
        Add an edge between the i-th and j-th points,
        if not in the list already
        """
        if (i, j) in edges or (j, i) in edges:
            # already added
            assert (j, i) in edges, "Can't go twice over same directed edge right?"
            if only_outer:
                # if both neighboring triangles are in shape, it's not a boundary edge
                edges.remove((j, i))
            return
        edges.add((i, j))

    tri = Delaunay(points)
    edges = set()
    # Loop over triangles:
    # ia, ib, ic = indices of corner points of the triangle
    for ia, ib, ic in tri.vertices:
        pa = points[ia]
        pb = points[ib]
        pc = points[ic]
        # Computing radius of triangle circumcircle
        # www.mathalino.com/reviewer/derivation-of-formulas/derivation-of-formula-for-radius-of-circumcircle
        a = np.sqrt((pa[0] - pb[0]) ** 2 + (pa[1] - pb[1]) ** 2)
        b = np.sqrt((pb[0] - pc[0]) ** 2 + (pb[1] - pc[1]) ** 2)
        c = np.sqrt((pc[0] - pa[0]) ** 2 + (pc[1] - pa[1]) ** 2)
        s = (a + b + c) / 2.0
        area = np.sqrt(s * (s - a) * (s - b) * (s - c))
        circum_r = a * b * c / (4.0 * area)
        if circum_r < alpha:
            add_edge(edges, ia, ib)
            add_edge(edges, ib, ic)
            add_edge(edges, ic, ia)
    return edges


def hash_numpy_array(array: np.ndarray) -> float:
    """
    Function to hash a numpy array.
    """
    
    # return hashlib.md5(array.data.tobytes()).hexdigest() # This is non deterministic.
    
    return np.sum(array ** 2)



#!/usr/bin/python

# Implementation of general tight-binding model based on the Slater-Koser approximation to parametrize
# the hopping amplitudes in term of the angles and orbitals between atoms.
# Requires an input file with the data regarding the model (dimensionality, number of atoms in motif, orbitals,
# interaction values)

import argparse
import sys
from tightbinder.fileparse import parse_config_file
from tightbinder.topology import calculate_wannier_centre_flow, calculate_z2_invariant, plot_wannier_centre_flow
import time
from tightbinder.models import SlaterKoster
import matplotlib.pyplot as plt

plt.rcParams.update({
    "text.usetex": True,
    "font.family": "serif"
})


def main():
    """ Main routine """
    # ------------------------------ Argument & file parsing ------------------------------
    parser = argparse.ArgumentParser()
    parser.add_argument('file', help='input file for tight-binding model')
    parser.add_argument('-v', '--verbose', action='store_true', help='verbose output')
    parser.add_argument('-cv', '--crystalview', action='store_true', help='3D view of the crystal')
    parser.add_argument('-z2', '--z2invariant', action='store_true', help='Compute Z2 invariant')
    parser.add_argument('-z2p', '--z2plot', action='store_true', help='Plot Wannier charge centers used for Z2 computation')
    parser.add_argument('-e', '--export', action='store_true', help='Export model Hamiltonian to file')
    args = parser.parse_args()

    try:
        file = open(args.file, 'r')
    except IOError:
        print('Error: Input file does not exist')
        sys.exit(1)

    configuration = parse_config_file(file)
    if 'Radius' in configuration:
        model = SlaterKoster(configuration, mode='radius', r=configuration['Radius'])
    else:
        model = SlaterKoster(configuration)

    # Possible system modifications before initialize_hamiltonian
    model.initialize_hamiltonian()
    print(len(model.bonds))

    if args.crystalview:
        model.visualize()

    if args.z2invariant:
        wcc = calculate_wannier_centre_flow(model, 20)
        z2  = calculate_z2_invariant(wcc)
        print(f'Z2 invariant is: {z2}')

        if args.z2plot:
            plot_wannier_centre_flow(wcc, show_midpoints=True)

    if args.export:
        modelfile = args.file
        # Remove extension if present and blanks
        modelfile = modelfile.replace(".txt", '').strip() + ".model"
        model.export_model(modelfile, fmt="%10.6f")
        print(f"Written model to file {modelfile}")

    labels = configuration['High symmetry points']
    kpoints = model.high_symmetry_path(configuration['Mesh'][0], labels)

    results = model.solve(kpoints)
    results.plot_along_path(labels, title=f'{model.system_name}')

    plt.show()


if __name__ == "__main__":
    initial_time = time.time()
    main()
    print(f'Elapsed time: {time.time() - initial_time}s')



from tightbinder.models import SlaterKoster, AmorphousSlaterKoster
from tightbinder.fileparse import parse_config_file
from tightbinder.disorder import amorphize
from tightbinder.utils import hash_numpy_array
import numpy as np
import math
import matplotlib.pyplot as plt


"""
This file constains tests for the main classes of the library, i.e. SlaterKoster and AmorphousSlaterKoster.
This also serves to test the base classes System and Crystal, which are inherited by the former.
All tests are functional, as they intend to check specific functionalities of the library and not 
specific routines individually.
"""

def test_configuration_parsing():
    """
    Function to test the parsing of a configuration file.
    """

    file = open("./examples/hBN.txt", "r")
    config = parse_config_file(file)

    assert config['Dimensionality'] == 2
    
    assert len(config['Bravais lattice']) == 2
    assert config['Bravais lattice'][0] == [2.16506, 1.25, 0.0]
    assert config['Bravais lattice'][1] == [2.16506, -1.25, 0.0]
    
    assert config['Species'] == 2
    
    assert len(config['Motif']) == 2
    assert config['Motif'][0] == [0, 0, 0, 0]
    assert config['Motif'][1] == [1.443376, 0, 0, 1]
    
    assert len(config['Filling']) == 2
    assert config['Filling'][0] == 0.5
    assert config['Filling'][1] == 0.5
    
    assert len(config['Orbitals']) == 2
    assert config['Orbitals'][0] == ['s']
    assert config['Orbitals'][1] == ['s']
    
    assert len(config['Onsite energy']) == 2
    assert config['Onsite energy'][0] == [3.625]
    assert config['Onsite energy'][0] == [3.625]
    
    assert len(config['SK amplitudes'].keys()) == 1
    assert config['SK amplitudes']['1']['01'] == [-2.3, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    
    assert config['Spin'] == False 
    
    assert len(config['Spin-orbit coupling']) == 2
    assert config['Spin-orbit coupling'][0] == 0
    assert config['Spin-orbit coupling'][1] == 0
    
    assert len(config['Mesh']) == 2
    assert config['Mesh'][0] == 100
    assert config['Mesh'][1] == 100
    
    assert config['High symmetry points'] == ['G', "K", 'M', "G"]
    

def build_slaterkoster() -> SlaterKoster:
    """
    Function to build a SlaterKoster model from the configuration file.
    """
    
    file = open("./examples/hBN.txt", "r")
    config = parse_config_file(file)
    
    model = SlaterKoster(config)
    
    return model


def test_sk_initialization():
    
    model = build_slaterkoster()
    

def test_supercell():
    """
    Routine to test the supercell method from System with a basic SlaterKoster model.
    """
    
    model = build_slaterkoster().supercell(n1=2, n2=2)
    
    assert model.boundary == "PBC"
    
    assert model.natoms == 8    
    assert np.allclose(model.bravais_lattice[0], [4.330128, 2.5, 0.0])
    assert np.allclose(model.bravais_lattice[1], [4.330128, -2.5, 0.0])   
    
    motif_hash = hash_numpy_array(model.motif)
    
    assert math.isclose(motif_hash, 99.833119893184)
        

def test_reduce():
    """
    Tests reduce method of System for generation of supercells with open boundary conditions.
    """
    
    model = build_slaterkoster().reduce(n1=2, n2=2)
    
    assert model.boundary == "OBC"    
    assert model.natoms == 8    
    
    motif_hash = hash_numpy_array(model.motif)
        
    assert math.isclose(motif_hash, 99.833119893184)
    
    
def test_add_bonds():
    """
    Tests the method add_bonds from System.
    """
    
    model = build_slaterkoster()
    model.add_bonds([0, 0], [0, 1], model.bravais_lattice)
    
    assert model.bonds[0][0] == 0
    assert model.bonds[0][1] == 0
    assert np.allclose(model.bonds[0][2], model.bravais_lattice[0])
    
    assert model.bonds[1][0] == 0
    assert model.bonds[1][1] == 1
    assert np.allclose(model.bonds[1][2], model.bravais_lattice[1])
        
    model.add_bond(1, 1, model.bravais_lattice[0] + model.bravais_lattice[1], '2')
    
    assert model.bonds[2][0] == 1
    assert model.bonds[2][1] == 1
    assert np.allclose(model.bonds[2][2], model.bravais_lattice[0] + model.bravais_lattice[1])
    assert model.bonds[2][3] == '2'
    

def test_neighbour_distances():
    """
    Tests that neighbour distances are computed correctly.
    """    
    
    model = build_slaterkoster().supercell(n1=4, n2=4)
    model.initialize_hamiltonian()
    
    distances = model.compute_neighbour_distances(5)
        
    assert np.allclose(distances, [1.44, 2.5,  2.89, 3.82, 4.33])
    

def test_neighbours():
    """
    Tests the method find_neighbours from System.
    """
    
    file = open("./examples/hBN.txt", "r")
    config = parse_config_file(file)
    
    model = SlaterKoster(config)
    model.find_neighbours()
    
    assert model.bonds[0][0] == 0
    assert model.bonds[0][1] == 1
    assert np.allclose(model.bonds[0][2], np.array([-2.16506, -1.25   ,  0.     ]))
    
    assert model.bonds[1][0] == 0
    assert model.bonds[1][1] == 1
    assert np.allclose(model.bonds[1][2], np.array([-2.16506,  1.25   ,  0.     ]))
    
    assert model.bonds[2][0] == 0
    assert model.bonds[2][1] == 1
    assert np.allclose(model.bonds[2][2], np.array([0., 0., 0.]))
    
    assert model.bonds[3][0] == 1
    assert model.bonds[3][1] == 0
    assert np.allclose(model.bonds[3][2], np.array([0., 0., 0.]))
    
    assert model.bonds[4][0] == 1
    assert model.bonds[4][1] == 0
    assert np.allclose(model.bonds[4][2], np.array([2.16506, -1.25   , 0.     ]))
    
    assert model.bonds[5][0] == 1
    assert model.bonds[5][1] == 0
    assert np.allclose(model.bonds[5][2], np.array([2.16506, 1.25   , 0.     ]))
    

def test_high_symmetry_path():
    """
    Tests that the high symmetry path specified is correctly generated.
    """
    
    model = build_slaterkoster()
    model.initialize_hamiltonian()
    
    kpoints = model.high_symmetry_path(10, ['G', 'K', 'M', 'G'])
    
    kpoints_hash = hash_numpy_array(np.array(kpoints))
        
    assert math.isclose(kpoints_hash, 12.243192500457077)
        
        
def test_bloch_hamiltonian():
    """
    Method to test the generation of the Bloch Hamiltonian from the results of the diagonalization.
    """
    
    model = build_slaterkoster()
    model.initialize_hamiltonian()
    
    kpoints = model.high_symmetry_path(10, ['G', 'K', 'M', 'G'])
    results = model.solve(kpoints)
    
    energies_hash = hash_numpy_array(results.eigen_energy)
        
    assert math.isclose(energies_hash, 667.8478374529238)
    

def test_fermi_energy():
    """
    Tests that the Fermi energy is computed properly.
    """
    
    model = build_slaterkoster()
    model.initialize_hamiltonian()
    
    kpoints = model.high_symmetry_path(100, ['G', 'K', 'M', 'G'])
    results = model.solve(kpoints)
    
    fermi_energy = results.calculate_fermi_energy(model.filling)
    
    assert math.isclose(fermi_energy, -3.625)
    
    
def test_fermi_energy_supercell():
    """
    Tests that the Fermi energy is computed properly when working with supercells.
    """
    
    model = build_slaterkoster().supercell(n1=2, n2=2)
    model.initialize_hamiltonian()
    
    labels = ['G', 'K', 'M', 'G']
    kpoints = model.high_symmetry_path(100, labels)
    results = model.solve(kpoints)
    
    fermi_energy = results.calculate_fermi_energy(model.filling)
    
    assert math.isclose(fermi_energy, -3.625)
    

def test_rescale_bands_to_fermi_energy():
    """
    Tests method to set the Fermi energy to zero.
    """
    
    model = build_slaterkoster()
    model.initialize_hamiltonian()
    
    kpoints = model.high_symmetry_path(100, ['G', 'K', 'M', 'G'])
    results = model.solve(kpoints)
    
    results.rescale_bands_to_fermi_level()
    fermi_energy = results.calculate_fermi_energy(model.filling)
    
    assert math.isclose(fermi_energy, 0)
    
    
def test_soc_bands():
    """
    Tests that SOC is correctly implemented.
    """
    
    file = open("./examples/Bi111.txt", "r")
    config = parse_config_file(file)
    
    model = SlaterKoster(config)
    model.initialize_hamiltonian()
    
    kpoints = model.high_symmetry_path(10, ['G', 'K', 'M', 'G'])
    results = model.solve(kpoints)
        
    results.rescale_bands_to_fermi_level()
    
    energies_hash = hash_numpy_array(results.eigen_energy)
        
    assert math.isclose(energies_hash, 5155.652053621002)
    

def test_edge_bands():
    """
    Tests that edge bands appear on the band structure of the ribbon of a topological insulator.
    """
    
    file = open("./examples/Bi111.txt", "r")
    config = parse_config_file(file)
    
    model = SlaterKoster(config).reduce(n1=5)
    model.initialize_hamiltonian()
    
    kpoints = model.high_symmetry_path(10, ['K', 'G', 'K'])
    results = model.solve(kpoints)
        
    results.rescale_bands_to_fermi_level()
    
    energies_hash = hash_numpy_array(results.eigen_energy)
            
    assert math.isclose(energies_hash, 29453.85050171607)
    

def test_amorphous_slater_koster():
    """
    Tests that the AmorphousSlaterKoster model can be initialized, disordered and the band structure can be obtained correctly.
    """
    
    np.random.seed(1)
    
    # Parameters of the calculation
    ncells = 4
    disorder = 0.1

    # Parse configuration file
    file = open("./examples/Bi111.txt", "r")
    config = parse_config_file(file)

    # Init. model and construct supercell
    first_neighbour_distance = np.linalg.norm(config["Motif"][1][:3])
    cutoff = first_neighbour_distance * 1.4
    
    model = AmorphousSlaterKoster(config, r=cutoff).supercell(n1=ncells, n2=ncells)
    model.decay_amplitude = 1

    model = amorphize(model, disorder)
    model.initialize_hamiltonian(override_bond_lengths=True)

    results = model.solve()
    
    energies_hash = hash_numpy_array(results.eigen_energy)
                
    assert math.isclose(energies_hash, 8595.611419662699)
    


""" 
Implementation of System class.

The System class extendes the Crystal class, and it 
has the basic functionality for the other models to derive from it.
"""

from __future__ import annotations
from typing import Tuple, Union, List
from .crystal import Crystal
import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg
import matplotlib.pyplot as plt
import sys
from itertools import product



class System(Crystal):
    """ 
    System class provides all basic functionality to create any type of model. This class
    serves as a base class for the actual models to inherit from. By itself it does not solve
    any system as it lacks specific implementation of the Hamiltonian, which has to be provided
    in the actual model implementation.

    :param system_name: Name of the system. Defaults to None.
    :param bravais_lattice: List with bravais lattice basis. Defaults to None.
    :param motif: List of atoms with their positions and their species. Defaults to None.
    :param crystal: Crystal object. Defaults to None.
    """
    
    def __init__(self, system_name: str = None, bravais_lattice: Union[list, np.ndarray] = None, 
                 motif: Union[list, np.ndarray] = None, crystal: Crystal = None):
        if crystal is None:
            super().__init__(bravais_lattice, motif)
        else:
            assert type(crystal) == Crystal
            if bravais_lattice is not None or motif is not None:
                print("Warning: Provided values for Bravais lattice or motif are disregarded" +
                      "in presence of crystal object")
            super().__init__(crystal.bravais_lattice, crystal.motif)

        self.system_name = system_name
        self._norbitals = None
        self._basisdim = None
        self._filling = None
        self._boundary = None
        self.bonds = []
        self._unit_cell_list = None
        self.hamiltonian = None
        self.first_neighbour_distance = None
        self._matrix_type = "dense"
        # TODO add flag for initialize_hamiltonian

    # ####################################################################################
    # #################################### Properties ####################################
    # ####################################################################################

    @property
    def norbitals(self):
        return self._norbitals

    @norbitals.setter
    def norbitals(self, norbitals):
        assert type(norbitals) == list
        self._norbitals = norbitals

    @property
    def filling(self):
        return self._filling

    @filling.setter
    def filling(self, filling):
        if filling < 0:
            print("Error: filling attribute must be between 0 and norbitals")
        else:
            self._filling = filling

    @property
    def boundary(self):
        return self._boundary

    @boundary.setter
    def boundary(self, boundary):
        if boundary not in ["PBC", "OBC"]:
            print('Error: Incorrect boundary option')
            sys.exit(1)
        self._boundary = boundary

    @property
    def basisdim(self):
        return self._basisdim

    @basisdim.setter
    def basisdim(self, basisdim):
        if basisdim < self.natoms:
            raise ValueError(f"basisdim has to be at least equal to the number of atoms (motif, {self.natoms})")
        self._basisdim = basisdim

    @property
    def matrix_type(self):
        return self._matrix_type

    @matrix_type.setter
    def matrix_type(self, type):
        if type not in ["dense", "sparse"]:
            raise ValueError("Invalid matrix type. Must be either dense or sparse.")
        self._matrix_type = type
    
    @property
    def unit_cell_list(self):
        if self._unit_cell_list is None:
            return []
        for cell in self._unit_cell_list:
            return self._unit_cell_list



    # ####################################################################################
    # ################################# Bonds/Neighbours #################################
    # ####################################################################################

    def add_bond(self, initial: int, final: int, cell: Union[list, tuple, np.ndarray] = (0., 0., 0.), neigh: str = '1') -> None:
        """ 
        Method to add a hopping between two atoms of the motif.
        NB: The hopping has a specified direction, from initial to final. The HSince the final
        Hamiltonian is computed taking into account hermiticity, it is not necessary to specify the hopping
        in the other direction. (TODO: Change this, no longer working like this).

        :param initial: Index of initial atom of the bond.
        :param final: Index of final atom of the bond.
        :param cell: Array with cell containing the final atom. Defaults to [0., 0., 0.].
        :param neigh: String to denote if the bond corresponds to first neighbour, second, etc. 
            Defaults to '1'.
        """

        hopping_info = [initial, final, cell, neigh]
        self.bonds.append(hopping_info)

    def add_bonds(self, initial: List[int], final: List[int], cells: Union[list, np.ndarray] = None, neigh: List[str] = None) -> None:
        """ 
        Same method as add_hopping but we input a list of hoppings at once.

        :param initial: List of indices of initial atom of the bond.
        :param final: List of indices of final atom of the bond.
        :param cell:  List of cell vectors containing the final atom. Defaults to None, which
            corresponds to the origin cell.
        :param neigh: List of strings to denote if the bonds corresponds to first neighbour, second, etc respectively.
            Defaults to None, which corresponds to all being '1'.
        """

        if len(initial) != len(final):
            raise ValueError("Initial and final lists do not have same size")
        if cells is None:
            cells = np.zeros([len(initial), 3])
        if neigh is None:
            neigh = ['1']*len(initial)
        else:
            self.boundary = "PBC"
            cells = np.array(cells)
        for bond in zip(initial, final, cells, neigh):
            self.add_bond(bond[0], bond[1], bond[2], bond[3])

    def compute_neighbour_distances(self, nni: int) -> List[float]:
        """ 
        Method to compute neighbour distance up to neighbours nni.

        :param nni: Maximum neighbour distance to consider. nni=5 mean that it
            will find first, second up to fifth neighbours distances.
        :return: List of neighbour distances.
        """

        if self.boundary == "OBC":
            near_cells = np.array([[0.0, 0.0, 0.0]])
        else:
            near_cells = generate_near_cells(self.bravais_lattice, nni)

        neigh_distance = np.array([])
        atoms = np.kron(np.array(self.motif)[:, :3], np.ones((near_cells.shape[0], 1))) + np.kron(np.ones((len(self.motif), 1)), near_cells)
        for reference_atom in self.motif:
            distance = np.linalg.norm(atoms - reference_atom[:3], axis=1)
            distance = np.unique(distance)[1:] # Remove 0 from distances
            neigh_distance = np.concatenate((neigh_distance, distance))
        
        neigh_distance = neigh_distance.round(2)
        neigh_distance = np.unique(neigh_distance)[:nni]

        return neigh_distance

    def find_first_neighbours(self) -> list:
        """ 
        Method to find the first neighbours of the atoms of the motif.
        :return: List of the corresponding bonds. 
        """

        if self.boundary == "OBC":
            near_cells = np.array([[0.0, 0.0, 0.0]])
        else:
            near_cells = generate_near_cells(self.bravais_lattice)
        first_neigh_distance = self.compute_first_neighbour_distance(near_cells)
        eps = 1E-2
        atoms = np.copy(self.motif)
        bonds = []
        for n, reference_atom in enumerate(self.motif):
            for cell in near_cells:
                distance = np.linalg.norm(atoms[:, :3] + cell - reference_atom[:3], axis=1)
                neigh_atoms_indices_max = np.where(distance <= first_neigh_distance + eps)[0]
                neigh_atoms_indices_min = np.where(first_neigh_distance - eps < distance)[0]
                neigh_atoms_indices = np.intersect1d(neigh_atoms_indices_max, neigh_atoms_indices_min)
                for i in neigh_atoms_indices:
                    bonds.append([n, i, cell])
        
        return bonds

    def find_neighbours(self, mode: str = "minimal", nn: int = 1, r: float = None) -> None:
        """ 
        From a list of atoms (motif), it sets the list of bonds within the solid.
        I.e.: Atom list -> List of neighbours/atom.
        By default it will look for the minimal distance between atoms to determine first neighbours.
        For amorphous systems the option radius is available to determine neighbours within a given radius R.
        Boundary conditions can also be set, either PBC (default) or OBC.

        :param mode: Search mode, can be either 'minimal' or 'radius'. Defaults to 'minimal'.
        :param nn: Next neighbour, used to specify up to which neighbour there are hoppings if mode='minimal'
        :param r: Value for radius sphere to detect neighbours if mode='radius' 
        """

        if self.bonds is not None:
            self.bonds = []
        eps = 1E-2
        if mode is "radius":
            if r is None:
                raise Exception("Error: Search mode is radius but no r given, exiting...")
            # TODO: Check if this is necessary
            # min_cell_size = np.min(np.linalg.norm(self.bravais_lattice, axis=1))
            # nn = int(r // min_cell_size + 1)

        elif mode is "minimal" and r is not None:
            print("Search mode is minimal but a radius was given (will not be used)")

        # Prepare unit cells to loop over depending on boundary conditions
        if self.boundary == "OBC":
            near_cells = np.array([[0.0, 0.0, 0.0]])
        else:
            near_cells = generate_near_cells(self.bravais_lattice, nn)

        # Determine neighbour distances up to nn
        neigh_distances = self.compute_neighbour_distances(nn)
        self.first_neighbour_distance = neigh_distances[0]
        # print(f"Neighbour distances: {neigh_distances}")
        if mode == "radius" and r < self.first_neighbour_distance:
            print("Warning: Radius smaller than first neighbour distance")

        # Look for neighbours
        # First in mode='minimal'
        atoms = np.copy(self.motif)
        index = 0
        if mode == 'minimal':
            for n, reference_atom in enumerate(self.motif):
                for cell in near_cells:
                    distance = np.linalg.norm(atoms[:, :3] + cell - reference_atom[:3], axis=1)
                    for nn, nn_distance in enumerate(neigh_distances):
                        neigh_atoms_indices_max = np.where(distance <= nn_distance + eps)[0]
                        neigh_atoms_indices_min = np.where(nn_distance - eps < distance)[0]

                        neigh_atoms_indices = np.intersect1d(neigh_atoms_indices_max, neigh_atoms_indices_min)
                        
                        for i in neigh_atoms_indices:
                            self.add_bond(n, i, cell, str(nn + 1))


        # Then in mode='radius'
        else:
            for n, reference_atom in enumerate(self.motif):
                for cell in near_cells:
                    distance = np.linalg.norm(atoms[:, :3] + cell - reference_atom[:3], axis=1)
                    neigh_atoms_indices_max = np.where(distance <= r + eps)[0]
                    neigh_atoms_indices_min = np.where(eps < distance)[0]
                    neigh_atoms_indices = np.intersect1d(neigh_atoms_indices_max, neigh_atoms_indices_min)
                    for i in neigh_atoms_indices:
                        self.add_bond(n, i, cell)

        print("Done")

    def find_disconnected_atoms(self) -> List[int]:
        """ 
        Method to find which atoms do not have neighbours, i.e. they are disconnected from
        the lattice.

        :return: List of indices of disconnected atoms.
        """
        
        all_bonds = self.__reconstruct_all_bonds()
        initial_atoms = [initial_atom for initial_atom, _, _ in all_bonds]
        disconnected_atoms = [atom for atom in range(self.natoms) if atom not in initial_atoms]

        return disconnected_atoms

    def remove_disconnected_atoms(self):
        """ 
        Method to remove disconnected atoms from the motif. Based on the method remove_atoms.
        """
        
        disconnected_atoms = self.find_disconnected_atoms()
        self.remove_atoms(disconnected_atoms)

    def __reconstruct_all_bonds(self) -> list:
        """ 
        Method to obtain all neighbours for all atoms since the find_neighbours method
        only computes one-way hoppings (i.e. i->j and not j->i).
        NB: Not necessary in principle since we now compute all explicitly.
        
        :return: List with all bonds.
        """
        
        all_bonds = self.bonds[:]
        for bond in self.bonds:
            initial, final, cell, nn = bond
            all_bonds.append([final, initial, cell, nn])

        return all_bonds

    @staticmethod
    def atom_coordination_number(index: int, bonds: list) -> int:
        """ 
        Method to find the coordination number for a specific atom in the crystal. 
        
        :param index: Index of atom whose coordination number we want to compute.
        :param bonds: List of indices within the solid.
        :return: Number of neighbours.
        """
        
        neighbours = np.where(bonds[:, 0] == index)[0]
        return len(neighbours)

    def coordination_number(self) -> Tuple[float, List[int]]:
        """ 
        Method to find the coordination number of the solid. Returns coordination number of solid,
        and a list with the different individual coordinations numbers found within the solid. 
        
        :return: Coordination number, and list of all individual coordination numbers found.
        """
        
        # bonds = self.__reconstruct_all_bonds()
        bonds = np.array(self.bonds, dtype=object)
        coordination = 0
        coordination_list = []
        for index in range(self.natoms):
            coordination_number = self.atom_coordination_number(index, bonds)
            coordination += coordination_number
            if coordination_number not in coordination_list:
                coordination_list.append(coordination_number)

        coordination_list = np.sort(coordination_list)
        coordination /= self.natoms
        return coordination, coordination_list

    def find_lowest_coordination_atoms(self, include_first_neighbours: bool = False) -> List[int]:
        """
        Routine to determine the atoms of the solid with the lowest coordination atoms.
        TODO: Define parameter to specify storing atoms with coordinations lower than
        one specified.
        
        :param include_first_neighbours: Includes in the list of atoms, also the first
            neighbours of those found originally. Defaults to False.
        :return: List of indices of atoms.
        """

        # bonds = self.__reconstruct_all_bonds()
        bonds = np.array(self.bonds, dtype=object)
        if bonds.size == 0:
            raise RuntimeError("Model must be initialized first")

        coordination, coordination_list = self.coordination_number()
        atoms = []
        for index in range(self.natoms):
            atom_coordination = self.atom_coordination_number(index, bonds)
            if atom_coordination < coordination_list[-1]:
                atoms.append(index)

        if include_first_neighbours:
            for atom in np.copy(atoms):
                neighbours = np.where(bonds[:, 0] == atom)[0]
                for index in neighbours:
                    neighbour = bonds[index][1]
                    if neighbour not in atoms:
                        atoms.append(neighbour)
        return atoms

    def compute_first_neighbour_distance(self, near_cells: Union[list, np.ndarray] = None) -> float:
        """
        Routine to compute the first neighbour distance.
        DEPRECATED since there is compute_neighbour_distances which works up to n neighbour distances.
        
        :param near_cells: List of unit cells to loop over. Defaults to None.
        :return: First neighbour distance.
        """

        if near_cells is None:
            near_cells = generate_near_cells(self.bravais_lattice)

        neigh_distance = 1E100
        fixed_atom = self.motif[0][:3]
        for atom, cell in product(self.motif, near_cells):
            distance = np.linalg.norm(atom[:3] + cell - fixed_atom)
            if 1E-4 < distance < neigh_distance:
                neigh_distance = distance

        self.first_neighbour_distance = neigh_distance

        return neigh_distance

    def _determine_connected_unit_cells(self) -> None:
        """ 
        Method to calculate which unit cells connect with the origin from the neighbour list. 
        """

        unit_cell_list = [[0.0, 0.0, 0.0]]
        if self.boundary == "PBC":
            for bond in self.bonds:
                unit_cell = list(bond[2])
                if unit_cell not in unit_cell_list:
                    unit_cell_list.append(unit_cell)

        self._unit_cell_list = unit_cell_list

    def identify_boundary(self, alpha: float = 0.4, ndim: int = 2, verbose: bool = False, corner_atoms: bool = False) -> np.ndarray:
        """ 
        Method to obtain the indices of the outermost atoms of the motif, taking into account
        boundary conditions. NB: Currently works only in 2D

        :param alpha: This parameters tunes the shape of the boundary. For alpha=0,
            the boundary is a convex hull; as alpha increases the boundary becomes more
            concave. Default value is 0.4
        :param ndim: Dimension of the expected boundary. Defaults to 2.
        :param verbose: Boolean parameter to print indices of boundary atoms. Defaults to False.
        :param cornes_atoms: Boolean to toggle detection of atoms that might be shared by the 
            periodic boundary and the open boundary.
        :return: Indices of atoms in the boundary.
        :raises AssertionError: Crystal dimension different from two raises error. Also 
            raises if bonds are not computed.
        """

        # Requires having computed the bonds previously
        if not self.bonds:
            raise AssertionError("Bonds must be computed first")
        bonds = self.bonds

        edges = self.identify_motif_edges(alpha, ndim)
        if not edges:
            import copy
            # If there are no edges, is because the system is effectively unidimensional, 
            # meaning that the Delaunay triangulation fails. Then consider to supercell to
            # be able to triangulate.

            bigger_system = copy.deepcopy(self).supercell(n1=2)
            bigger_system.initialize_hamiltonian()
            bonds = bigger_system.bonds

            edges = bigger_system.identify_motif_edges(alpha, ndim)

        boundary_atoms = [edge[i] for edge in edges for i in range(2)]
        boundary_atoms = np.unique(boundary_atoms)

        # Remove boundary atoms from the periodic boundary
        # First compute atoms in the periodic boundary
        periodic_boundary_atoms = []
        for bond in bonds:
            cell = bond[2]
            if np.linalg.norm(cell) != 0:
                periodic_boundary_atoms.append(bond[0])
                periodic_boundary_atoms.append(bond[1])
        periodic_boundary_atoms = np.unique(periodic_boundary_atoms)      
        
        boundary_atoms = np.setdiff1d(boundary_atoms, periodic_boundary_atoms)

        # Detect atoms in the corner between periodic and finite boundary
        if corner_atoms:
            corner_atoms = []
            for atom in boundary_atoms:
                for bond in bonds:
                    if bond[0] == atom and bond[1] in periodic_boundary_atoms:
                        corner_atoms.append(bond[1])

            corner_atoms = np.unique(corner_atoms)
            boundary_atoms = np.concatenate([boundary_atoms, corner_atoms])

        atoms = []
        for atom in boundary_atoms:
            if atom in range(len(self.motif)):
                atoms.append(atom)
        atoms = np.unique(atoms)

        if verbose:
            print(f"Atoms in boundary: {atoms}")

        if len(atoms) == 0:
            print("Warning: No boundary atoms found")

        return np.array(atoms, dtype=int)
    
    def plot_wireframe(self, ax: plt.Axes = None, linewidth: float = 1, alpha: float = 0.5, color="black") -> None:
        """
        Method to plot a wireframe of the system according to the detected bonds.

        :param ax: Matplotlib axes object. If None, a new Axis object is created.
        :param linewidth: Linewidth of the wireframe. Defaults to 1.
        :param alpha: Sets wireframe transparency. Defaults to 0.5
        :param color: Wireframe color. Defaults to black.
        """

        if not self.bonds:
            raise LookupError("plot_wireframe requires having computed the bonds of the system.")

        if ax is None:
            fig, ax = plt.subplots(1, 1)
        
        for n, bond in enumerate(self.bonds):

            x0, y0 = self.motif[bond[0], :2]
            xneigh, yneigh = self.motif[bond[1], :2]
            ax.plot([x0, xneigh], [y0, yneigh], "-", linewidth=linewidth, alpha=alpha, color=color)


    # ####################################################################################
    # ############################### System modifications ###############################
    # ####################################################################################

    def supercell(self, update: bool = True, **ncells: dict) -> System:
        """ 
        Routine to generate a supercell for a system using the number of cells
        along each Bravais vectors specified in ncells.

        :param update: Boolean parameter to update all of Crystal class atributes.
            Defaults to True, only set to false when used inside other routines (reduce)
        :param dict: Specify number of cells in some direction: n1, n2 and/or n3.
        :return: Modified System object. 
        """

        if len(ncells) == 0:
            print("Error: Reduce method must be called with at least one parameter (nx, ny or nz), exiting...")
            sys.exit(1)
        key_to_index = {"n1": 0, "n2": 1, "n3": 2}
        for key in ncells.keys():
            if key not in ["n1", "n2", "n3"]:
                print("Error: Invalid input (must be n1, n2 or n3), exiting...")
                sys.exit(1)
            if key_to_index[key] > self.ndim:
                print(f"Error: Axis {key} to reduce along not present (higher than system dimension)")

            new_motif = self.motif
            for n in range(1, ncells[key]):
                motif_copy_displaced = np.copy(self.motif)
                motif_copy_displaced[:, :3] += n * self.bravais_lattice[key_to_index[key]]
                new_motif = np.append(new_motif, motif_copy_displaced, axis=0)

            # Update Bravais lattice and motif
            self.bravais_lattice[key_to_index[key]] *= ncells[key]
            self.motif = new_motif
        
        if update:
            self.update()

        return self

    def reduce(self, **ncells: dict) -> System:
        """ 
        Routine to reduce the dimensionality of the System object along the specified
        directions, by repeating unit cells along those directions until a given size
        (number of unit cells) is reached. Thus we make the original system finite along those
        directions.

        :param ncells: Number of cells of direction(s) along which we want to reduce the
            dimensionality of the system, n1, n2 and/or n3.
        :return: Modified System object.
        """

        key_to_index = {"n1": 0, "n2": 1, "n3": 2}
        self.supercell(update=False, **ncells)
        indices = [index for index in list(range(self.ndim))
                   if index not in [key_to_index[key] for key in ncells.keys()]]
        if not indices:
            self.bravais_lattice = None
            self.boundary = "OBC"
        else:
            self.bravais_lattice = self.bravais_lattice[indices]

        return self

    def ribbon(self, width: int, orientation: str = "horizontal", periodic: bool = False) -> System:
        """ 
        Routine to generate a ribbon for a 2D crystal. It is designed to create rectangular ribbons,
        based on a rectangular lattice. Therefore there must exist a combination of basis vectors such that
        we can obtain a rectangular supercell. Otherwise the method returns an error. 

        :param width: Width of the ribbon (number of cells).
        :param orientation: 'horizontal' or 'vertical'. Defaults to 'horizontal'.
        :param periodic: Boolean to keep the ribbon periodic along the direction it is built. Defauts to False.
            Turning this one would correspond to a nanotube instead of nanoribbon (finite).
        :return: Modified System object.
        """
        
        if orientation not in ["horizontal", "vertical"]:
            raise ValueError("Wrong orientation provided, must be either vertical or horizontal, exiting...")
        
        if self.ndim != 2:
            print(f"Ribbons can not be generated for {self.ndim}D structures")
            sys.exit(1)
        else:
            mesh_points = []
            for _ in range(self.ndim):
                mesh_points.append(list(range(-1, 2)))

            combinations = np.array(np.meshgrid(*mesh_points)).T.reshape(-1, self.ndim)
            rectangular_basis = np.zeros([2, 3])
            all_possible_atoms = []
            motif = []

            # Calculate rectangular lattice basis vectors
            for n, m in combinations:
                vector = n * self.bravais_lattice[0] + m * self.bravais_lattice[1]
                for atom in self.motif:
                    new_atom = np.copy(atom)
                    new_atom[:3] += vector
                    all_possible_atoms.append(new_atom)

                if n == m == 0:
                    continue
                vector_angle = np.arctan2(vector[1], vector[0])

                if abs(vector_angle - 0.0) < 1E-10:
                    rectangular_basis[0, :] = vector
                elif abs(vector_angle - np.pi/2) < 1E-10:
                    rectangular_basis[1, :] = vector

            if not np.linalg.norm(rectangular_basis, axis=1).any():
                print("Error: Could not generate a rectangular lattice, use reduce method directly. Exiting...")
                sys.exit(1)

            # Calculate new motif
            for i, atom in enumerate(all_possible_atoms):
                if (0 <= atom[0] < np.linalg.norm(rectangular_basis[0, :])) and \
                   (0 <= atom[1] < np.linalg.norm(rectangular_basis[1, :])):
                    motif.append(atom)

            # Update system attributes
            self.bravais_lattice = rectangular_basis
            self.motif = motif

            # Reduce system dimensionality
            if not periodic:
                if orientation == "horizontal":
                    self.reduce(n1=width)
                else:
                    self.reduce(n2=width)
            else:
                if orientation == "horizontal":
                    self.supercell(n1=width)
                else:
                    self.supercell(n2=width)


            return self
        
    def extend_copy(self, vector: List[list, np.ndarray], n: int = 1) -> System:
        """
        Method to extend the system by attaching a copy of the motif displaced by some input vector.

        :param vector: Vector by which we displace the copy of the system.
        :param n: Number of copies to attach. The n-th copy will be displaced by n*vector. Defaults to 1.
        :return: Extended system.
        """

        displaced_motif = np.copy(self.motif)
        for _ in range(n):
            displaced_motif[:, :3] += np.array(vector)
            self.motif = np.concatenate((self.motif, displaced_motif))  

        return self
    
    def rotate(self, angle: float = 90) -> System:
        """
        Method to rotate the whole system in the xy plane by some angle.

        :param angle: Value of rotation angle in degrees. Defaults to 90 degrees.
        :return: Rotated system.
        """

        rotation_matrix = np.array([[np.cos(np.pi*angle/180), -np.sin(np.pi*angle/180), 0],
                                    [np.sin(np.pi*angle/180),  np.cos(np.pi*angle/180), 0],
                                    [                      0,                        0, 1]])

        for n in range(self.natoms):
            self.motif[n, :3] = rotation_matrix @ self.motif[n, :3]

        rotated_bravais = np.zeros(self.bravais_lattice.shape)
        for n in range(self.ndim):
            rotated_bravais[n] = rotation_matrix @ self.bravais_lattice[n]
        self.bravais_lattice = rotated_bravais

        return self

    def restrict_lattice2rectangle(self) -> None:
        """
        TO BE IMPLEMENTED YET (is it really necessary?).
        """
        
        atoms = []
        mesh_points = []
        for i in range(self.ndim):
            mesh_points.append(list(range(-1, 2)))

        combinations = np.array(np.meshgrid(*mesh_points)).T.reshape(-1, self.ndim)
        for n, m in combinations:
            vector = n*self.bravais_lattice[0] + m*self.bravais_lattice[1]
            for position in self.motif:
                atom_position = vector + position


    # ####################################################################################
    # ################################### Hamiltonian ####################################
    # ####################################################################################

    def initialize_hamiltonian(self):
        """ 
        Generic implementation of initialization of the Hamiltonian. To be overwritten
        by specific implementations of System. 
        """
        
        print("initialize_hamiltonian has to be implemented by child class")

    def hamiltonian_k(self, k: Union[list, np.ndarray]) -> np.ndarray:
        """ 
        Generic implementation of hamiltonian_k H(k). To be overwritten
        by specific implementations of System. 
        """
        
        print("Has to be implemented by child class")

    def solve(self, kpoints: list = None, neigval: int = None):
        """ 
        Diagonalize the Hamiltonian to obtain the band structure and the eigenstates. 
        
        :param kpoints: List of kpoints where we want to diagonalize the Bloch Hamiltonian.
            Defaults to None, which corresponds to [[0., 0., 0.]] for both PBC and OBC.
        :param neigval: Number of eigenvalues to compute if using sparse matrices.
            Defaults to None, which in turn is set to the filling.
        :return: Spectrum object.
        """

        # Import result here to avoid circular import
        from . import result
        
        if kpoints is None:
            kpoints = np.array([[0., 0., 0.]])
        if neigval is None and self.matrix_type == "sparse":
            neigval = int(self.filling)

        nk = len(kpoints)
        eigen_energy = np.zeros([self.basisdim, nk])
        eigen_states = []

        if self.matrix_type == "dense":
            solver = lambda h: np.linalg.eigh(h)
        else:
            solver = lambda h: scipy.sparse.linalg.eigsh(h, k=neigval)

        for n, k in enumerate(kpoints):
            hamiltonian = self.hamiltonian_k(k)
            results = solver(hamiltonian)
            eigen_energy[:, n] = results[0]
            eigen_states.append(results[1])

        return result.Spectrum(eigen_energy, np.array(eigen_states), kpoints, self)


class FrozenClass:
    """ 
    Class to enforce immutable attributes. Used for predefined models.
    """
    
    _is_frozen = False

    def __setattr__(self, key, value):
        if self._is_frozen:
            raise Exception("Predefined model can not be modified")
        super().__setattr__(key, value)

    def _freeze(self):
        self._is_frozen = True


def generate_all_combinations(ndim: int, n: int = 1) -> np.ndarray:
    """ 
    Auxiliary routine to generate an array of combinations of possible neighbouring
    unit cells. 
    
    :param ndim: Number of axis to make the combinations.
    :param int: Values on each axis go from [-n, n]. Defaults to 1.
    :return: Array with vectors of length ndim and combinations.
    """
    
    mesh_points = []
    for i in range(ndim):
        mesh_points.append(list(range(-n, n+1)))
    mesh_points = np.array(np.meshgrid(*mesh_points)).T.reshape(-1, ndim)

    return mesh_points


def generate_half_combinations(ndim: int) -> np.ndarray:
    """ 
    Auxiliary routine to generate an array of combinations of possible neighbouring
    unit cells. NB: It only generates half of them, since we are going to use hermiticity to
    generate the Hamiltonian. 
    
    :param ndim: Number of axis to make the combinations.
    :return: Half the combinations.
    """
    
    all_combinations = generate_all_combinations(ndim)

    # Eliminate vectors that are the inverse of others
    points = []
    for point in all_combinations:
        if list(-point) not in points:
            points.append(list(point))

    return np.array(points)


def generate_near_cells(bravais_lattice: Union[list, np.ndarray], n: int = 1, half: bool = False) -> np.ndarray:
    """ 
    Auxiliary routine to generate the Bravais vectors corresponding to unit cells
    neighbouring the origin one. NB: It only generates half of them, since we are going to use hermiticity to
    generate the Hamiltonian. 
    
    :param bravais_lattice: Basis vectors of the Bravais lattice.
    :param n: Values on each axis go from [-n, n]. Defaults to 1.
    :param half: Generate only half the combinations. Defaults to False.
    :return: Array with combinations of Bravais basis vectors.
    """
    
    if bravais_lattice is None:
        return np.array([[0., 0., 0.]])
        
    ndim = len(bravais_lattice)
    if not half:
        mesh_points = generate_all_combinations(ndim, n)
    else:
        mesh_points = generate_half_combinations(ndim, n)
    near_cells = np.zeros([len(mesh_points), 3])
    for n, coefficients in enumerate(mesh_points):
        cell_vector = np.array([0.0, 0.0, 0.0])
        for i, coefficient in enumerate(coefficients):
            cell_vector += (coefficient * bravais_lattice[i])
        near_cells[n, :] = cell_vector

    return near_cells






""" 
Implementation of the Crystal class, responsible of storing all the 
information relative to the Bravais lattice and the reciprocal lattice,
such as high symmetry points, producing the k point mesh or visualizing
the crystal itself.
"""

import numpy as np
from scipy.spatial import ConvexHull
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from numpy.linalg import LinAlgError
import math
import sys
from typing import Union, List
from .utils import generate_basis_combinations, alpha_shape_2d


# --------------- Constants ---------------
PI = 3.141592653589793238
EPS = 0.001


# --------------- Routines ---------------
class Crystal:
    """
    Implementation of Crystal class. Defaults to square lattice if no input is given.

    :param bravais_lattice: List with the bravais vectors 
    :param motif: List with positions and chemical species of atoms of the motif
    """

    def __init__(self, bravais_lattice: Union[list, np.ndarray] = None, motif: Union[list, np.ndarray] = None) -> None:
        self.group = None
        self.reciprocal_basis = None
        self.high_symmetry_points = None

        self._ndim = None
        self.natoms = None
        self.unit_cell_area = None
        self._bravais_lattice = None
        self.bravais_lattice = bravais_lattice
        self._motif = None
        self.motif = motif

    @property
    def ndim(self) -> int:
        return self._ndim

    @ndim.setter
    def ndim(self, ndim: int) -> None:
        assert type(ndim) == int
        if ndim > 3 or ndim < (-1):
            print("Incorrect dimension for system, exiting...")
            sys.exit(1)
        self._ndim = ndim

    @property
    def bravais_lattice(self) -> np.ndarray:
        return self._bravais_lattice

    @bravais_lattice.setter
    def bravais_lattice(self, bravais_lattice: Union[list, np.ndarray]) -> None:
        if bravais_lattice is None:
            self._bravais_lattice = None
            self._ndim = 0
        else:
            assert type(bravais_lattice) == list or type(bravais_lattice) == np.ndarray
            bravais_lattice = np.array(bravais_lattice)
            if bravais_lattice.shape[1] != 3:
                raise Exception("Vectors must have three components")
            self._bravais_lattice = bravais_lattice
            self._ndim = len(bravais_lattice)
            self.update()

    @property
    def motif(self) -> Union[list, np.ndarray]:
        return self._motif

    @motif.setter
    def motif(self, motif: Union[list, np.ndarray]) -> None:
        if motif is None:
            self._motif = None
        else:
            assert type(motif) == list or type(motif) == np.ndarray
            for atom in motif:
                if len(atom) != 4:
                    raise Exception("Each position must have three components")
            self._motif = np.array(motif)
            self.natoms = len(motif)

    def add_atom(self, position: Union[list, np.ndarray], species: int = 0) -> None:
        """
        Method to add one atom from a numbered species into a specific position.

        :param position: Array with atom coordinaes. Must be three-dimensional.
        :param species: Index of corresponding chemical species. Defaults to 0. 
        """

        assert type(position) == list or type(position) == np.ndarray
        if len(position) != 3:
            raise Exception("Vector must have three components")
        
        position = np.concatenate([position, [species]])

        if type(self.motif) == list:
            self.motif.append(position)
        elif type(self.motif) == np.ndarray:
            self.motif = np.concatenate([self.motif, [position]], axis=0)

    def add_atoms(self, atoms: Union[list, np.ndarray], species: list = None) -> None:
        """ 
        Method to add a list of atoms of specified species at some positions.
        Built on top of method add_atom.
        
        :param atoms: List of length n, containing the positions of each atom.
        :param species: List with the indices of the chemical species. If None, all default to zero. 
        """

        if species is None:
            species = np.zeros(len(atoms))
        for n, atom in enumerate(atoms):
            self.add_atom(atom, species[n])

    def remove_atom(self, index: int) -> None:
        """ 
        Method to remove the atom at the specified index from the motif. Note
        that this method does not remove existing bonds corresponding to the removed atom.

        :param index: Index of atom to be removed. 
        """

        if type(self.motif) == list:
            self.motif.pop(index)
            self.natoms -= 1
        elif type(self.motif) == np.ndarray:
            self.motif = np.delete(self.motif, index, 0)

    def remove_atoms(self, indices: list) -> None:
        """ 
        Method to remove the atoms specified in a list with all their indices. 
        Built upon the method remove_atom.
        
        :param indices: List with the indices of the atoms to be removed. 
        """

        indices = np.sort(indices)
        for n, index in enumerate(indices):
            self.remove_atom(index - n)

    def update(self) -> None:
        """ 
        Routine to initialize or update the intrinsic attributes of the Crystal class whenever
        the Bravais lattice is changed. 
        """

        self.crystallographic_group()
        self.reciprocal_lattice()
        self.determine_high_symmetry_points()
        self.compute_unit_cell_area()

    def compute_unit_cell_area(self) -> float:
        """ 
        Method to compute the area of the unit cell. 
        TODO: Adapt for 1D and 3D 

        :return: Value of unit cell area.
        """
        
        area = 0.0
        if self.ndim == 2:
            area = np.linalg.norm(np.cross(self.bravais_lattice[0], self.bravais_lattice[1]))
            self.unit_cell_area = area

        return area

    def crystallographic_group(self) -> None:
        """ 
        Determines the crystallographic group associated to the material from
        the configuration file. NOTE: Only the group associated to the Bravais lattice,
        reduced symmetry due to presence of motif is not taken into account. Its purpose is to
        provide a path in k-space to plot the band structure.
        Workflow is the following: First determine length and angles between basis vectors.
        Then we separate by dimensionality: first we check angles, and then length to
        determine the crystal group. 
        """

        basis_norm = [np.linalg.norm(vector) for vector in self.bravais_lattice]
        group = ''

        basis_angles = []
        for i in range(self.ndim):
            for j in range(self.ndim):
                if j <= i: continue
                v1 = self.bravais_lattice[i]
                v2 = self.bravais_lattice[j]
                basis_angles.append(math.acos(np.dot(v1, v2)/(np.linalg.norm(v1)*np.linalg.norm(v2))))

        if self.ndim == 2:
            if abs(basis_angles[0] - PI / 2) < EPS:
                if abs(basis_norm[0] - basis_norm[1]) < EPS:
                    group += 'Square'
                else:
                    group += 'Rectangular'

            elif (abs(basis_angles[0] - PI / 3) < EPS or
                  abs(basis_angles[0] - 2 * PI / 3) < EPS) and abs(basis_norm[0] - basis_norm[1]) < EPS:
                group += 'Hexagonal'

            else:
                if abs(basis_norm[0] - basis_norm[1]) < EPS:
                    group += 'Centered rectangular'
                else:
                    group += 'Oblique'

        elif self.ndim == 3:
            if (basis_angles[0] - PI/2) < EPS and (basis_angles[1] - PI/2) < EPS and (basis_angles[2] - PI/2) < EPS:
                if abs(basis_norm[0] - basis_norm[1]) < EPS and abs(basis_norm[0] - basis_norm[2]) < EPS and abs(basis_norm[1] - basis_norm[2]) < EPS:
                    group += 'Cube'

            else:
                print('Work in progress')

        self.group = group

    def reciprocal_lattice(self) -> None:
        """ 
        Routine to compute the reciprocal lattice basis vectors from
        the Bravais lattice basis. The algorithm is based on the fact that
        a_i\dot b_j=2PI\delta_ij, which can be written as a linear system of
        equations to solve for b_j. Resulting vectors have 3 components independently
        of the dimension of the vector space they span.
        """

        if self.bravais_lattice is None:
            self.reciprocal_basis = None
            return

        reciprocal_basis = np.zeros([self.ndim, 3])
        coefficient_matrix = np.array(self.bravais_lattice)

        if self.ndim == 1:
            reciprocal_basis = 2*PI*coefficient_matrix/(np.linalg.norm(coefficient_matrix)**2)

        else:
            coefficient_matrix = coefficient_matrix[:, :self.ndim]
            for i in range(self.ndim):
                coefficient_vector = np.zeros(self.ndim)
                coefficient_vector[i] = 2*PI
                try:
                    reciprocal_vector = np.linalg.solve(coefficient_matrix, coefficient_vector)
                except LinAlgError:
                    print('Error: Bravais lattice basis is not linear independent')
                    sys.exit(1)

                reciprocal_basis[i, 0:self.ndim] = reciprocal_vector

        self.reciprocal_basis = reciprocal_basis

    def brillouin_zone_mesh(self, mesh: list) -> np.ndarray:
        """ 
        Routine to compute a mesh of the first Brillouin zone using the
        Monkhorst-Pack algorithm.

        :param mesh: List with the number of kpoints along each axis.
        :return: Array with all kpoints, nk x 3 
        """

        if len(mesh) != self.ndim:
            print('Error: Mesh does not match dimension of the system')
            sys.exit(1)

        kpoints = []
        mesh_points = []
        for i in range(self.ndim):
            mesh_points.append(list(range(0, mesh[i])))

        mesh_points = np.array(np.meshgrid(*mesh_points)).T.reshape(-1, self.ndim)

        for point in mesh_points:
            kpoint = 0
            for i in range(self.ndim):
                kpoint += (2.*point[i] - mesh[i])/(2*mesh[i])*self.reciprocal_basis[i]
            kpoints.append(kpoint)

        return np.array(kpoints)

    def determine_high_symmetry_points(self) -> None:
        """ 
        Routine to compute high symmetry points depending on the dimension of the system.
        These symmetry points will be used to plot the band structure along the main
        reciprocal paths of the system (irreducible BZ). Returns a dictionary with pairs
        {letter denoting high symmetry point: its coordinates} 
        """

        norm = np.linalg.norm(self.reciprocal_basis[0])
        special_points = {r"$\Gamma$": np.array([0., 0., 0.])}
        if self.ndim == 1:
            special_points.update({'K': self.reciprocal_basis[0]/2})

        elif self.ndim == 2:
            special_points.update({'M': self.reciprocal_basis[0]/2})
            if self.group == 'Square':
                special_points.update({'K': self.reciprocal_basis[0]/2 + self.reciprocal_basis[1]/2})

            elif self.group == 'Rectangular':
                special_points.update({'X': special_points['M']})
                special_points.update({'Y': self.reciprocal_basis[1]/2})
                special_points.update({'S': self.reciprocal_basis[0]/2 + self.reciprocal_basis[1]/2})

            elif self.group == 'Hexagonal':
                # special_points.update({'K',: reciprocal_basis[0]/3 + reciprocal_basis[1]/3})
                special_points.update({'K': norm/math.sqrt(3)*(self.reciprocal_basis[0]/2 - self.reciprocal_basis[1]/2)/(
                                             np.linalg.norm(self.reciprocal_basis[0]/2 - self.reciprocal_basis[1]/2))})

            elif self.group == 'Oblique' or self.group == 'Centered rectangular':
                if np.linalg.norm(self.bravais_lattice[0]) > np.linalg.norm(self.bravais_lattice[1]):
                    a = np.linalg.norm(self.bravais_lattice[0])
                    b = np.linalg.norm(self.bravais_lattice[1])
                else:
                    a = np.linalg.norm(self.bravais_lattice[1])
                    b = np.linalg.norm(self.bravais_lattice[0])
                gamma = math.acos(np.dot(self.bravais_lattice[0], 
                                         self.bravais_lattice[1])/(a*b))
                eta = (1 - b*math.cos(gamma)/a)/(2*(math.sin(gamma)**2))
                nu = 0.5 - eta*a*math.cos(gamma)/b

                special_points.update({'X': special_points['M']})
                special_points.update({'Y': self.reciprocal_basis[1]/2})
                special_points.update({'C': self.reciprocal_basis[0]/2 + self.reciprocal_basis[1]/2})
                special_points.update({'H': eta*self.reciprocal_basis[0] + (1 - nu)*self.reciprocal_basis[1]})
                special_points.update({r"$H_1$": (1- eta)*self.reciprocal_basis[0] + nu*self.reciprocal_basis[1]})

            else:
                print('High symmetry points not implemented yet')

        else:
            if self.group == 'Cube':
                special_points.update({'X': self.reciprocal_basis[0]/2})
                special_points.update({'M': self.reciprocal_basis[0]/2 + self.reciprocal_basis[1]/2})
                special_points.update({'R': self.reciprocal_basis[0]/2 + self.reciprocal_basis[1]/2 +
                                            self.reciprocal_basis[2]/2})

            else:
                print('High symmetry points not implemented yet')

        self.high_symmetry_points = special_points

    def __reorder_high_symmetry_points(self, labels: list) -> None:
        """ 
        Routine to reorder the high symmetry points according to a given vector of labels
        for later representation.
        DEPRECATED with the dictionary update 
        """
        
        points = []
        for label in labels:
            appended = False
            for point in self.high_symmetry_points:
                if appended:
                    continue
                if label == point[0]:
                    points.append(point)
                    appended = True

        self.high_symmetry_points = points

    def __strip_labels_from_high_symmetry_points(self) -> None:
        """ 
        Routine to output both labels and array corresponding to high symmetry points
        DEPRECATED with the dictionary update 
        """

        for n, point in enumerate(self.high_symmetry_points):
            self.high_symmetry_points[n] = point[1]

    def high_symmetry_path(self, nk: int, points: List[str]) -> list:
        """ 
        Routine to generate a path in reciprocal space along the high symmetry points (HSPs) indicated.
        The HSPs must be within those corresponding to the associated symmetry group, otherwise the method will throw an error.

        :param nk: Number of kpoints in the path. They are evenly distributed between HSPs.
        :param points: List of labels of desired HSPs.
        :return: List of kpoints following the path specified with the labels. 
        """

        # Transform possible Gamma point to latex
        for n, point in enumerate(points):
            if point == "G":
                points[n] = r"$\Gamma$"
            elif point == "H1":
                points[n] = r"$H_1$"

        kpoints = []
        number_of_points = len(points)
        interval_mesh = int(nk/(number_of_points - 1))
        previous_point = self.high_symmetry_points[points[0]]
        for point in points[1:]:
            next_point = self.high_symmetry_points[point]
            kpoints += list(np.linspace(previous_point, next_point, interval_mesh, endpoint=False))
            previous_point = next_point
        kpoints.append(self.high_symmetry_points[points[-1]])

        return kpoints

    def identify_motif_edges(self, alpha: float = 0.4, ndim: int = 2) -> np.ndarray:
        """
        Method to obtain the indices of the outermost atoms of the motif.
        Note that this method returns exclusively the outermost atoms, independently
        of the boundary conditions. To take into account boundary conditions, one should
        call identify_boundary() from System.
        NB: Currently works only in 2D

        :param alpha: This parameters tunes the shape of the boundary. For alpha=0,
            the boundary is a convex hull; as alpha increases the boundary becomes more
            concave. Default value is 0.4
        :param ndim: Dimension of the boundary. If the system lies on a plane, the boundary
            is two-dimensional, so ndim = 2 (default value).
        :return: Indices of atoms in the boundary.
        :raises AssertionError: Crystal dimension different from two raises error.
        :raises ValueError: Crystal motif must have at least three atoms. Raised by Delaunay class.
        """

        if ndim == 2:
            points = np.array(self.motif)[:, :2]
            points[:, 0] = (points[:, 0] - np.min(points[:, 0]))/np.max(points[:, 0])
            points[:, 1] = (points[:, 1] - np.min(points[:, 1]))/np.max(points[:, 1])

            edges = alpha_shape_2d(points, alpha)

            return edges

        else:
            raise AssertionError("Currently working only for ndim=2")

    def identify_convex_hull(self, loop: int = 6) -> np.ndarray:
        """
        Deprecated in favour of identify_motif_edges. 
        Method to obtain the atoms that correspond to the edges of the system.
        This method uses directly the ConvexHull method from Scipy to identify the outmost points
        of the motif. Compute the ConvexHull three times, each time removing the points detected in the
        previous iteration.

        NB: If given 3d points, they must not be coplanar for the algorithm
        to work. If they are, then the points must be given as 2d 

        :param loop: Number of times the ConvexHull is applied, to increase number of points in the edge.
        :return: Indices of atoms belonging to edges of motif.
        """

        atoms_coordinates = np.copy(self.motif[:, :2])
        hull = ConvexHull(atoms_coordinates)
        edge_indices = hull.vertices
        max_x = np.max(atoms_coordinates[:, 0])
        min_x = np.min(atoms_coordinates[:, 0])
        max_y = np.max(atoms_coordinates[:, 1])
        min_y = np.min(atoms_coordinates[:, 1])
        midpoint = np.array([(max_x - min_x)/2, (max_y - min_y)/2])
        for index in edge_indices:
            atoms_coordinates[index, :] = midpoint

        # Repeat this procedure more times
        for _ in range(loop):
            hull = ConvexHull(atoms_coordinates)
            edge_indices = np.concatenate((edge_indices, hull.vertices))
            for index in hull.vertices:
                atoms_coordinates[index, :] = midpoint

        return edge_indices

    def crystal_center(self) -> np.ndarray:
        """ 
        Method to compute the geometric center of the crystal from the positions of the atoms
        of the motif. Calculated as the average of all atomic positions.

        :return: Numpy array with the cartesian coordinates of the center 
        """

        center = np.sum(self.motif[:, :3], axis=0)/self.natoms
        return center


    def plot_crystal(self, cell_number: int = 1, crystal_name: str = '') -> None:
        """ 
        Method to visualize the crystalline structure (Bravais lattice + motif).

        :param cell_number: Number of cells appended in each direction with respect to the origin. Defaults to 1.
        :param crystal_name: Name of the crystal. Defaults to the empty string. 
        """

        bravais_lattice = np.array(self.bravais_lattice)
        motif = np.array(self.motif)

        fig = plt.figure()
        ax = Axes3D(fig)
        bravais_vectors_mesh = []
        for i in range(self.ndim):
            bravais_vectors_mesh.append(list(range(-cell_number, cell_number + 1)))
        bravais_vectors_mesh = np.array(np.meshgrid(*bravais_vectors_mesh)).T.reshape(-1, self.ndim)

        bravais_vectors = np.zeros([len(bravais_vectors_mesh[:, 0]), 3])
        for n, coefs in enumerate(bravais_vectors_mesh):
            for i, coef in enumerate(coefs):
                bravais_vectors[n, :] += coef * bravais_lattice[i]

        all_atoms_list = np.zeros([len(bravais_vectors_mesh)*len(motif), 3])
        iterator = 0
        for vector in bravais_vectors:
            for atom in motif:
                all_atoms_list[iterator, :] = atom[:3] + vector
                iterator += 1

        [min_axis, max_axis] = [np.min(all_atoms_list), np.max(all_atoms_list)]

        for atom in all_atoms_list:
            ax.scatter(atom[0], atom[1], atom[2], color='y', s=50)

        ax.set_xlabel('x (A)')
        ax.set_ylabel('y (A)')
        ax.set_zlabel('z (A)')
        ax.set_title(crystal_name + ' crystal')
        ax.set_xlim3d(min_axis, max_axis)
        ax.set_ylim3d(min_axis, max_axis)
        ax.set_zlim3d(min_axis, max_axis)
        plt.axis('off')
        plt.show()

    def plot_2d_crystal(self, ax = None, fontsize: int = 10, size: int = 10) -> None:
        """ 
        Method to visualize the crystalline structure (Bravais lattice + motif) in two dimensions.
        If the crystal is 3D, it will plot only the x, y coordinates.
        
        :param ax: Axis to plot the projection of the crystal. If None, creates one.
        :param fontsize: Size of labels. Defaults to 10.
        :param size: Size of markers. Defaults to 10.
        """

        if ax is None:
            fig, ax = plt.subplots(1, 1)

        [min_axis, max_axis] = [np.min(self.motif), np.max(self.motif)]

        for atom in self.motif:
            ax.scatter(atom[0], atom[1], color='y', s=size)

        ax.set_xlabel('x (A)', fontsize=fontsize)
        ax.set_ylabel('y (A)', fontsize=fontsize)
        ax.set_xlim(min_axis, max_axis)
        ax.set_ylim(min_axis, max_axis)
        ax.axis('equal')
        ax.tick_params('both', labelsize=fontsize)

    def visualize(self) -> None:
        """ 
        Method to render a 3d visualization of the crystal using the self.vpython library. 
        """

        CrystalView(self).visualize()
        print("\nPress Ctrl+c to exit visualization or close the window.")
        while True:
            pass


class CrystalView:
    """ 
    The CrystalView class' aim is to wrap all the logic relative to the visualization of the crystal using the
    vpython library. This is, rendering the solid appropriately, and also rendering a minimal user interface to interact
    with the visualization. Note that typehints for VPython components have been deleted to avoid importing the 
    library beforehand.
    """

    def __init__(self, crystal: Crystal) -> None:

        self.bravais_lattice = crystal.bravais_lattice
        self.motif = np.array(crystal.motif)
        self.ndim = crystal.ndim
        self.atoms = []
        self.extra_atoms = []
        self.bonds = []
        self.extra_bonds = []
        self.atom_radius = None
        try:
            self.vp = __import__("vpython")
        except:
            raise AssertionError("visualize(): Requires having installed the library VPython.")

        try:
            self.edge_atoms = crystal.identify_boundary(corner_atoms=True)
        except AssertionError as e:
            print("CrystalView.visualize(): Unable to determine open boundary (too few atoms or inexistent with PBC)")

        try:
            self.neighbours = crystal.bonds
        except IndexError or AttributeError as e:
            print('visualize(): System must be initialized first')
            raise

        self.vp.scene.visible = False
        self.vp.scene.background = self.vp.color.black
        self.__compute_atom_radius()
        self.__create_atoms()
        self.__create_bonds()

        # Canvas attributes
        self.drag = False
        self.initial_position = None
        self.cell_vectors = None
        self.cell_boundary = None

    def __compute_atom_radius(self) -> None:
        """ 
        Private method to estimate an adequate radius for the atoms based on the distance between first neighbours. 
        """

        first_neigh_distance = 0
        for initial_atom, final_atom, cell, _ in self.neighbours:
            bond_length = np.linalg.norm(self.motif[initial_atom, :3] - self.motif[final_atom, :3] - cell)
            if (bond_length < first_neigh_distance) or first_neigh_distance == 0:
                first_neigh_distance = bond_length

        self.atom_radius = first_neigh_distance/7

    def __create_atoms(self) -> None:
        """ 
        Private method to initialize the atoms as spheres. 
        """

        # Origin cell
        mesh_points = generate_basis_combinations(self.ndim)
        color_list = [self.vp.color.yellow, self.vp.color.red, self.vp.color.white, self.vp.color.blue, self.vp.color.green]
        for position in self.motif:
            species = int(position[3])
            atom = self.vp.sphere(radius=self.atom_radius, color=color_list[species])
            atom.pos.x, atom.pos.y, atom.pos.z = position[:3]
            self.atoms.append(atom)

        # Super cell
        if self.bravais_lattice is not None:
            for point in mesh_points:
                unit_cell = np.array([0., 0., 0.])
                if np.linalg.norm(point) == 0:
                    continue
                for n in range(self.ndim):
                    unit_cell += point[n] * self.bravais_lattice[n]
                for position in self.motif:
                    species = int(position[3])
                    atom = self.vp.sphere(radius=self.atom_radius, color=color_list[species])
                    atom.visible = False
                    atom.pos.x, atom.pos.y, atom.pos.z = (position[:3] + unit_cell)
                    self.extra_atoms.append(atom)

    def __create_bonds(self) -> None:
        """ 
        Private method to initialized the bonds between atoms as solid lines. 
        """

        mesh_points = generate_basis_combinations(self.ndim)
        
        try:
            # Origin cell
            for initial_atom, final_atom, cell, nn in self.neighbours:
                unit_cell = self.vp.vector(*cell)
                bond = self.vp.curve(self.atoms[initial_atom].pos, self.atoms[final_atom].pos + unit_cell)
                self.bonds.append(bond)
                if np.linalg.norm(cell) > 1E-4:
                    continue
                    reverse_bond = self.vp.curve(self.atoms[final_atom].pos, self.atoms[initial_atom].pos - unit_cell)
                    self.bonds.append(reverse_bond)


            # Super cell
            for point in mesh_points:
                unit_cell = np.array(([0., 0., 0.]))
                for n in range(self.ndim):
                    unit_cell += point[n] * self.bravais_lattice[n]
                unit_cell = self.vp.vector(unit_cell[0], unit_cell[1], unit_cell[2])
                for bond in self.bonds:
                    supercell_bond = self.vp.curve(bond.point(0)["pos"] + unit_cell, bond.point(1)["pos"] + unit_cell)
                    supercell_bond.visible = False
                    self.extra_bonds.append(supercell_bond)

        except AttributeError:
            print("Warning: To visualize bonds, Viewer must receive an initialized System")

    def visualize(self) -> None:
        """ 
        Method to render the 3d visualization of the crystal. It also renders a GUI to interact
        with the visualization. 
        """

        self.vp.scene.visible = True
        self.vp.button(text="supercell", bind=self.__add_unit_cells)
        self.vp.button(text="primitive unit cell", bind=self.__remove_unit_cells)
        self.vp.button(text="remove bonds", bind=self.__remove_bonds)
        self.vp.button(text="show bonds", bind=self.__show_bonds)
        self.vp.button(text="draw cell boundary", bind=self.__draw_boundary)
        self.vp.button(text="remove cell boundary", bind=self.__remove_boundary)
        self.vp.button(text="highlight edge atoms", bind=self.__highlight_edge)
        self.vp.scene.bind("mousedown", self.__mousedown)
        self.vp.scene.bind("mouseup", self.__mouseup)

    def __mousedown(self, event) -> None:
        """ 
        Private method to detect mouse buttons press.

        :param event: vpython.event_return object. 
        """

        self.initial_position = self.vp.scene.mouse.pos
        self.drag = True

    def __mouseup(self, event) -> None:
        """ 
        Private method to detect mouse buttons release. 

        :param event: vpython.event_return object.
        """

        if self.drag:
            increment = self.vp.scene.mouse.pos - self.initial_position
            self.initial_position = self.vp.scene.mouse.pos
            self.vp.scene.camera.pos -= increment
        self.drag = False

    def __add_unit_cells(self, button) -> None:
        """ 
        Private method to render additional unit cells from the origin. 

        :param button: vpython.button object.
        """

        for atom in self.extra_atoms:
            atom.visible = True

    def __show_bonds(self, button) -> None:
        """ 
        Private method to render the bonds between atoms.

        :param button: vpython.button object.
        """

        for bond in self.bonds:
            bond.visible = True
        if self.extra_atoms[0].visible:
            for bond in self.extra_bonds:
                bond.visible = True

    def __remove_unit_cells(self, button) -> None:
        """ 
        Private method to derender the additional unit cells. 

        :param button: vpython.button object.
        """

        for atom in self.extra_atoms:
            atom.visible = False
        if self.extra_bonds[0].visible:
            for bond in self.extra_bonds:
                bond.visible = False

    def __remove_bonds(self, button) -> None:
        """ 
        Private method to derender the bonds between atoms. 

        :param button: vpython.button object.
        """

        for bond in self.bonds:
            bond.visible = False
        for bond in self.extra_bonds:
            bond.visible = False

    def __draw_boundary(self, button) -> None:
        """ 
        Private method to draw the boundary of the unit cell.
        TODO: Requires fixing. 

        :param button: vpython.button object.
        """

        self.cell_vectors = []
        mid_point = np.array([0., 0., 0.])
        size_vector = [0., 0., 0.]
        for n in range(self.ndim):
            mid_point += self.bravais_lattice[n] / 2
            size_vector[n] = self.bravais_lattice[n][n]
        self.cell_boundary = self.vp.box(pos=self.vp.vector(mid_point[0], mid_point[1], mid_point[2]),
                                    size=self.vp.vector(size_vector[0], size_vector[1], size_vector[2]),
                                    opacity=0.5)
        for basis_vector in self.bravais_lattice:
            unit_cell = self.vp.vector(basis_vector[0], basis_vector[1], basis_vector[2])
            vector = self.vp.arrow(pos=self.vp.vector(0, 0, 0),
                              axis=unit_cell,
                              color=self.vp.color.green,
                              shaftwidth=0.1)
            self.cell_vectors.append(vector)

    def __remove_boundary(self, button) -> None:
        """ 
        Private method to remove the boundary of the unit cell.

        :param button: vpython.button object. 
        """

        self.cell_boundary.visible = False
        for vector in self.cell_vectors:
            vector.visible = False

    def __highlight_edge(self, button) -> None:
        """ 
        Private method to highlight the edges of the atom of the solid. 

        :param button: vpython.button object.
        """

        if button.text == "highlight edge atoms":
            for atom_index in self.edge_atoms:
                self.atoms[atom_index].color = self.vp.color.green
                button.text = "dehighlight edge atoms"
        else:
            for atom_index in self.edge_atoms:
                self.atoms[atom_index].color = self.vp.color.yellow
                button.text = "highlight edge atoms"




"""
Modifications of the system such as strain or electric field.
"""

from shutil import ExecError
import numpy as np
from tightbinder.models import SlaterKoster
from tightbinder.system import System
from tightbinder.fileparse import mix_parameters

def apply_constant_electric_field(system: System, intensity: float, axis: str) -> System:
    """ 
    Routine to apply a constant electric field to a Hamiltonian, along the specified axis. Modifies
    on-site energies of the atoms of the system according to their positions.

    :param system: System object
    :param intensity: Amplitude of applied electric field
    :param axis: String
    :returns: System object 
    """
    
    pass

def strain(system: SlaterKoster, strain: float, n: int = 8) -> SlaterKoster:
    """ 
    Routine to simulate the effect of strain on a given system. Modifies all SK amplitudes following 
    a polynomial law to incorporate the effect of a lower bond length.

    :param system: SlaterKoster object
    :param strain: Value of strain, 0 <= strain < 1
    :param n: Power of
    :return: Strained SlaterKoster object. 
    """
    
    if strain >= 1 or strain < 0:
        raise ValueError('strain must be a number in [0, 1)')

    system.system_name = 'Strained' + system.system_name
    factor = (1 - strain)**(-n)
    for neighbour in system.configuration['SK amplitudes']:
        for species, coefs in system.configuration['SK amplitudes'][neighbour]:
            system.configuration['SK amplitudes'][neighbour][species] = [coef*factor for coef in coefs]

    return system

def saturate_bonds(system: SlaterKoster, onsite: float = 0, vss: float = 0, vsp: float = 0, edge_atoms: list = None) -> SlaterKoster:
    """ 
    Routine to saturate dangling bonds in materials by adding hydrogen atoms which are
    connected to the atoms at the edge of the system.
    The new species corresponds to hydrogen atoms (or equivalently some species with only s orbitals).
    Note that the specified hopping will be mixed with the hoppings provided for the other species.

    :param system: SlaterKoster model to saturate 
    :param onsite: Value of the onsite energy for H atoms.
    :param vss: Value of the SK amplitude for s-s orbitals. It will be mixed with the hopping of the
        species that participate. 
    :param vsp: Value of the SK amplitude for s-p orbitals. It will be mixed with the hopping of the
        species that participate. 
    :param edge_atoms: Optional list of atoms to which we connect the H atoms
    :return: Modified SlaterKoster object. 
    """

    if not system.hamiltonian:
        raise ExecError("System must be initialized first")

    if edge_atoms is None:
        edge_atoms = system.find_lowest_coordination_atoms()
    new_species = system.nspecies
    nn_distance = system.compute_first_neighbour_distance()
    motif_center = system.crystal_center()
    cell = [0., 0., 0.]
    for atom_index in edge_atoms:
        atom_position = system.motif[atom_index, :3]
        difference_position = atom_position - motif_center
        hydrogen_position = difference_position/np.linalg.norm(difference_position)*nn_distance + atom_position
        system.add_atom(hydrogen_position, new_species)
        system.add_bond(atom_index, system.natoms - 1, cell)
        system.add_bond(system.natoms - 1, atom_index, cell)
    
    if system.configuration["Spin"] == True:
        system.configuration["Filling"].append(1)
    else:
        system.configuration["Filling"].append(0.5)
    system.configuration["Species"] += 1
    system.configuration["Orbitals"].append(['s'])
    system.configuration["Onsite energy"].append([onsite])
    system.configuration["Spin-orbit coupling"].append(0)

    new_species = str(new_species) + str(new_species)
    amplitudes = [vss, vsp, 0, 0, 0, 0, 0, 0, 0, 0] # Hardcoded, maybe define some routine to perform this transformation
    system.configuration["SK amplitudes"]['1'][new_species] = amplitudes
    mix_parameters(system.configuration)

    return system
